#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

import re
import numpy as np
from functools import wraps, partial
import inspect
from copy import deepcopy
from collections import OrderedDict
import warnings


def validate_float(var_name,value,min_value=None,max_value=None):
    """
    Validates whether a variable can be used as a floating point number and whether it is within specified bounds
    If a value equals one of the bounds, the validation passes
    """
    try:
        float(value)
    except Exception as err:
        raise TypeError("%s (%s) is not a floating point number - %s" % (var_name,str(value),str(err)))
        
    if min_value!=None and value<min_value:
        raise ValueError("%s (%s) cannot be smaller than %s" % (var_name,str(value),str(min_value)))
        
    if max_value!=None and value>max_value:
        raise ValueError("%s (%s) cannot be greater than %s" % (var_name,str(value),str(max_value)))
    
    return True
    
def validate_integer(var_name,value,min_value=None,max_value=None):
    """
    Validates whether a variable can be used as an integer and whether it is within specified bounds
    If a value equals one of the bounds, the validation passes
    """
    try:
        if int(value)==value:
            pass
        else:
            raise TypeError("Value can be converted to integer (%s) but converted integer does not equal %s" % (str(int(value)),str(value)))
    except Exception as err:
        raise TypeError("%s (%s) is not an integer number - %s" % (var_name,str(value),str(err)))
        
    if min_value!=None and value<min_value:
        raise ValueError("%s (%s) cannot be smaller than %s" % (var_name,str(value),str(min_value)))
        
    if max_value!=None and value>max_value:
        raise ValueError("%s (%s) cannot be greater than %s" % (var_name,str(value),str(max_value)))
    
    return True

def validate_boolean(var_name,value):
    """
    Validates whether a variable can be used as a boolean
    """
    try:
        if bool(value)==value:
            pass
        else:
            raise TypeError("Value can be converted to boolean (%s) but converted boolean does not equal %s" % (str(bool(value)),str(value)))
    except Exception as err:
        raise TypeError("%s (%s) is not a boolean - %s" % (var_name,str(value),str(err)))
    
    return True
    
def validate_string(var_name,value,options=None,regex=None):
    """
    Validates whether a variable can be used as a string.
    The routine also allows checking whether the string is in a list of strings
    or whether it matches a specific regex pattern
    """
    try:
        if str(value)==value:
            pass
        else:
            raise TypeError("Value can be converted to string (%s) but converted string does not equal %s" % (str(value),str(value)))
    except Exception as err:
        raise TypeError("%s (%s) is not a string - %s" % (var_name,str(value),str(err)))
    
    if options!=None and value not in options:
        raise ValueError("%s (%s) not included in list of allowable strings (%s)" % (var_name,str(value),str(options)))
        
    if regex!=None and not bool(re.match(re.compile(regex), value)):
        raise ValueError("%s (%s) does not match the required string format (%s)" % (var_name,str(value),str(regex)))
    
    return True
   
def validate_list(var_name,value,elementtype=None,order=None,unique=None,empty_allowed=None):
    """
    Validates whether a list contains numbers. It allows checking whether these numbers are ascending or descending
    and whether non-unique values exist 
    """
    try:
        if type(value)==np.ndarray:
            value=list(value)

        if list(value)==value or tuple(value)==value:
            pass
        else:
            raise TypeError("Value can be converted to list (%s) but converted list does not equal %s" % (str(value),str(value)))
    except Exception as err:
        raise TypeError("%s (%s) is not a list or tuple - %s" % (var_name,str(value),str(err)))
    
    if elementtype!=None:
        try:
            for i,el in enumerate(value):
                if elementtype=="float":
                    validate_float(var_name,el)
                elif elementtype=="string":
                    validate_string(var_name,el)
                elif elementtype=="int":
                    validate_integer(var_name,el)
                elif elementtype=="boolean":
                    validate_boolean(var_name,el)
                else:
                    raise ValueError("Unspecified elementtype")
        except Exception as err:
            raise ValueError("Invalid element type for %s, %s required" % (str(el),elementtype))
    
    if order=='ascending':
        try:
            if sorted(value)==value and (np.NaN not in value):
                pass
            else:
                raise ValueError("List %s is not ascending" % str(value))
        except Exception as err:
            raise ValueError("%s" % str(err))
    elif order=='descending':
        try:
            if sorted(value)==list(reversed(value)) and (np.NaN not in value):
                pass
            else:
                raise ValueError("List %s is not descending" % str(value))
        except Exception as err:
            raise ValueError("%s" % str(err))
    elif order is None:
        pass # Nothing happens when order is not specified
    else:
        raise ValueError("Incorrect string for list order")
    
    if unique==True:
        if len(value) > len(set(value)):
            raise ValueError("%s (%s) contains non-unique elements" % (var_name,str(value)))
    elif unique is None or unique==False:
        pass # Nothing happens when unique is None or unspecified
    else:
        raise ValueError("Validation parameter unique must be boolean")
        
    if empty_allowed==False:
        if len(value)==0:
            raise ValueError("Empty lists are not allowed")
        
    return True

def map_args(method,var,*args,**kwargs):
    
    """
    Constructs a data structure with all parameters, their values and the validation parameters
    which need to be used during validation.
    
    :param method: The function for which validation will be applied
    :param var: The validation data structure, entered as argument of the function decorator
    :param args: function arguments
    :param kwargs: function keyword arguments
    
    :returns dictionary var_validation which is a copy of the validation data structure
             it is possible to override __min and __max arguments
    """
    try:
        # Construct a data structure with all function arguments, defaults are used
        # Remove self for validators applied to class methods
        parameter_names = [parameter.name for parameter in inspect.signature(method).parameters.values() \
                           if ((parameter.kind == parameter.POSITIONAL_OR_KEYWORD) and (parameter.name!='self'))]
        all_vars = OrderedDict.fromkeys(parameter_names)

        args = tuple(x for x in args if isinstance(x, (int, float, str, bool, complex, list, tuple, np.ndarray)))

        for parameter in inspect.signature(method).parameters.values():
            if str(parameter) != 'self':
                if not isinstance(parameter.default, type):
                    all_vars[parameter.name] = parameter.default

        for key, value in kwargs.items():
            if key in parameter_names:
                all_vars[key] = value

        for i, arg in enumerate(args):
            all_vars[list(all_vars.keys())[i]] = args[i]

        var_validation = deepcopy(var)
        
        for key in kwargs.keys():
            # Modification of min and max ranges with override
            # To be changed for not permanent override of min and max
            if key.endswith('__min'):
                var_validation[key.replace('__min','')]['min_value'] = kwargs[key]
            elif key.endswith('__max'):
                var_validation[key.replace('__max','')]['max_value'] = kwargs[key]
            # Bind the actual function arguments, this is required because the defaults are otherwise used 
            else:
                all_vars[key]=kwargs[key]
        
        # Add the value used at runtime to the validation data structure, except for raises_errors and validate
        # which are used elsewhere and not in the validation routine
        for key in all_vars.keys():
            try:
                var_validation[key]['value'] = all_vars[key]
            except:
                pass
        
        return var_validation
    except Exception as err:
        raise ValueError("Error during mapping of validation parameters to function parameters - %s" % str(err))

class ValidationDecorator(object):

    def __init__(self, argument):
        self.arg = argument

    def __call__(self, fn):
        @wraps(fn)
        def decorated(*args, **kwargs):
            validated = True
            errorstring = ""

            try:
                validate = kwargs['validate']
            except:
                validate = None

            try:
                fail_silently = kwargs['fail_silently']
            except:
                fail_silently = True

            if validate or validate is None:

                try:
                    var_validation = map_args(fn, self.arg, *args, **kwargs)

                    for v in var_validation.keys():

                        if var_validation[v]['type'] == 'float':
                            validate_float(v, var_validation[v]['value'],
                                           var_validation[v]['min_value'],
                                           var_validation[v]['max_value'])
                        elif var_validation[v]['type'] == 'int':
                            validate_integer(v, var_validation[v]['value'],
                                             var_validation[v]['min_value'],
                                             var_validation[v]['max_value'])
                        elif var_validation[v]['type'] == 'string':
                            validate_string(v, var_validation[v]['value'],
                                            options=var_validation[v]['options'],
                                            regex=var_validation[v]['regex'])
                        elif var_validation[v]['type'] == 'bool':
                            validate_boolean(v, var_validation[v]['value'])
                        elif var_validation[v]['type'] == 'list':
                            validate_list(v, var_validation[v]['value'],
                                          var_validation[v]['elementtype'],
                                          var_validation[v]['order'],
                                          var_validation[v]['unique'],
                                          var_validation[v]['empty_allowed'])

                except Exception as err:
                    validated = False
                    errorstring += str(err)
            else:
                pass  # No validation

            return fn(*args, validated=validated, errorstring=errorstring, **kwargs)

        return decorated


class Validator(object):
    """
    The Validator has the following features

        - Automatic handling of validation errors
        - Automatic handling of function output upon errors
        - Possibility to override the default validation dictionary with custom validation

    """

    def __init__(self, validationspec, outputonerrorspec):
        self.validationspec = validationspec
        self.outputonerror = outputonerrorspec

    def __call__(self, fn):
        @wraps(fn)
        def decorated(*args, **kwargs):

            try:
                validate = kwargs['validate']
            except:
                validate = None

            try:
                fail_silently = kwargs['fail_silently']
            except:
                fail_silently = True

            try:
                validation_params = kwargs['customvalidation']
            except:
                validation_params = self.validationspec

            try:
                output_for_errors = kwargs['customerroroutput']
            except:
                output_for_errors = self.outputonerror

            if validate or validate is None:
                # Execute validation
                try:
                    var_validation = map_args(fn, validation_params, *args, **kwargs)

                    for v in var_validation.keys():

                        if var_validation[v]['type'] == 'float':
                            validate_float(v, var_validation[v]['value'],
                                           var_validation[v]['min_value'],
                                           var_validation[v]['max_value'])
                        elif var_validation[v]['type'] == 'int':
                            validate_integer(v, var_validation[v]['value'],
                                             var_validation[v]['min_value'],
                                             var_validation[v]['max_value'])
                        elif var_validation[v]['type'] == 'string':
                            validate_string(v, var_validation[v]['value'],
                                            options=var_validation[v]['options'],
                                            regex=var_validation[v]['regex'])
                        elif var_validation[v]['type'] == 'bool':
                            validate_boolean(v, var_validation[v]['value'])
                        elif var_validation[v]['type'] == 'list':
                            validate_list(v, var_validation[v]['value'],
                                          var_validation[v]['elementtype'],
                                          var_validation[v]['order'],
                                          var_validation[v]['unique'],
                                          var_validation[v]['empty_allowed'])

                except:
                    raise
            else:
                # No validation
                pass

            try:
                result = fn(*args, **kwargs)
                return result
            except:
                if fail_silently:
                    return output_for_errors
                else:
                    raise

        return decorated

def check_layer_overlap(df, raise_error=True, z_from_key=None, z_to_key=None):
    """
    Checks possible overlap on a dataframe
    :param df: Dataframe with keys 'z from [m]' and 'z to [m]'. Other keys can be used but then the arguments `z_from_key` and `z_to_key` need to be provided.
    :param raise_error: Boolean determining whether an error needs to be raised or whether a warning is sufficient (default behaviour is to raise an error warning)
    :param z_from_key: Key for start depth of the layer
    :param z_to_key: Key for end depth of the layer
    :return: Default behaviour: raises a warning if there are overlaps or gaps.
    """
    # Reset the index first
    df.reset_index(drop=True, inplace=True)

    # Set keys for top and bottom depths of layers
    if z_from_key is None:
        z_from_key = "z from [m]"

    if z_to_key is None:
        z_to_key = "z to [m]"

    for i, row in df.iterrows():
        if i > 0:
            if row[z_from_key] > df.loc[i-1, z_to_key]:
                if raise_error:
                    raise ValueError("A gap exists between layer %i and %i" % (i-1, i))
                else:
                    warnings.warn("A gap exists between layer %i and %i" % (i-1, i))
            elif row[z_from_key] < df.loc[i-1, z_to_key]:
                if raise_error:
                    raise ValueError("Overlap exists between layer %i and %i" % (i-1, i))
                else:
                    warnings.warn("Overlap exists between layer %i and %i" % (i-1, i))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import warnings
from copy import deepcopy

# 3rd party packages
import numpy as np
from scipy.optimize import brentq
import pandas as pd
from plotly import tools
import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS
from plotly.offline import iplot

# Project imports
from groundhog.general.validation import Validator


STRESS_ELASTIC_ISOTROPIC = {
    'internal_pressure': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'farfield_pressure': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'borehole_radius': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'shear_modulus': {'type': 'float', 'min_value': 0.0, 'max_value': None},
}

STRESS_ELASTIC_ISOTROPIC_ERRORRETURN = {
    'radial stress [kPa]': np.nan,
    'tangential stress [kPa]': np.nan,
    'radial displacement [m]': np.nan,
}


@Validator(STRESS_ELASTIC_ISOTROPIC, STRESS_ELASTIC_ISOTROPIC_ERRORRETURN)
def stress_cylinder_elastic_isotropic(
        radius, internal_pressure, farfield_pressure, borehole_radius,
        shear_modulus=np.nan, **kwargs):
    """
    Calculates the radial and tangential stress around a cylindrical borehole under internal pressure, in a soil mass with isotropic virgin stress conditions in the given plane

    :param radius: Radius or radii at which to calculate the stresses (float or NumPy array) (:math:`r`) [:math:`m`] - Suggested range: radius >= 0.0
    :param internal_pressure: Internal pressure on the borehole (:math:`p`) [:math:`kPa`] - Suggested range: internal_pressure >= 0.0
    :param farfield_pressure: Far-field pressure, equal to the virgin horizontal stress (:math:`p_0`) [:math:`kPa`] - Suggested range: farfield_pressure >= 0.0
    :param borehole_radius: Radius of the borehole (:math:`a`) [:math:`m`] - Suggested range: borehole_radius >= 0.0
    :param shear_modulus: Shear modulus used to calculate the radial displacement (:math:`G`) [:math:`kPa`] - Suggested range: shear_modulus >= 0.0 (optional, default=np.nan). If unspecified, radial displacements are not calculated

    .. math::
        \\frac{d \\sigma_{r}}{dr} + \\frac{\\left(\\sigma_{r} - \\sigma_{\\theta} \\right)}{r} = 0

        \\sigma_r | _{r=a} = p \\\\
        \\sigma_r | _{r=a} = p_0

        \\sigma_{r} = p_0 + \\left( p - p_0 \\right) \\cdot \\left( \\frac{a}{r} \\right)^2

        \\sigma_{\\theta} = p_0 - \\left( p - p_0 \\right) \\cdot \\left( \\frac{a}{r} \\right)^2

    :returns: Dictionary with the following keys:

        - 'radial stress [kPa]': Radial stress at the specified radii (:math:`\\sigma_r`)  [:math:`kPa`]
        - 'tangential stress [kPa]': Tangential stress at the specified radii (:math:`\\sigma_{\\theta}`)  [:math:`kPa`]
        - 'radial displacement [m]': Radial displacement (calculated if shear modulus is specified) (:math:`u`)  [:math:`m`]

    .. figure:: images/stress_elastic_isotropic.png
        :figwidth: 500.0
        :width: 450.0
        :align: center

        Basic sketch of the cavity expansion problem

    Reference - Yu, H.-S., 2000. Cavity Expansion Methods in Geomechanics. Springer-Science+Business Media, B.V.

    """

    _radial_stress = farfield_pressure + (internal_pressure - farfield_pressure) * ((
        borehole_radius / radius) ** 2)
    _tangential_stress = farfield_pressure - (internal_pressure - farfield_pressure) * ((
        borehole_radius / radius) ** 2)
    if np.math.isnan(shear_modulus):
        try:
            _radial_displacement = list(map(lambda _x: np.nan, _radial_stress))
        except:
            _radial_displacement = np.nan
    else:
        _radial_displacement = ((internal_pressure - farfield_pressure) / (2 * shear_modulus)) * \
            radius * ((borehole_radius / radius) ** 2)

    return {
        'radial stress [kPa]': _radial_stress,
        'tangential stress [kPa]': _tangential_stress,
        'radial displacement [m]': _radial_displacement,
    }


EXPANSION_TRESCA_THICKSPHERE = {
    'undrained_shear_strength': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'internal_radius': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'external_radius': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'internal_pressure': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'external_pressure': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'youngs_modulus': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'poissons_ratio': {'type': 'float', 'min_value': 0.0, 'max_value': 0.5},
    'seed': {'type': 'int', 'min_value': None, 'max_value': None},
}

EXPANSION_TRESCA_THICKSPHERE_ERRORRETURN = {
    'elastic radii [m]': np.nan,
    'elastic_radial_stress [kPa]': np.nan,
    'elastic_tangential_stress [kPa]': np.nan,
    'elastic_radial_displacement [m]': np.nan,
    'yielding_pressure [kPa]': np.nan,
    'plastic_radius [m]': np.nan,
    'elastoplastic radii [m]': np.nan,
    'elastoplastic_radial_stress [kPa]': np.nan,
    'elastoplastic_tangential_stress [kPa]': np.nan,
    'expanded_radius [m]': np.nan,
}


@Validator(EXPANSION_TRESCA_THICKSPHERE, EXPANSION_TRESCA_THICKSPHERE_ERRORRETURN)
def expansion_tresca_thicksphere(
        undrained_shear_strength, internal_radius, external_radius, internal_pressure, external_pressure,
        youngs_modulus, poissons_ratio, seed=100, **kwargs):
    """
    Calculates the stresses for cavity expansion around a thick-walled sphere in Tresca material.

    The plastic radius is first calculated from the pressure boundary conditions. Using this plastic radius, the stresses in the elastic and plastic region are calculated.

    :param undrained_shear_strength: Undrained shear strength of the material surrounding the spherical cavity (:math:`S_u`) [:math:`kPa`] - Suggested range: undrained_shear_strength >= 0.0
    :param internal_radius: Initial internal radius of the spherical cavity (:math:`a_0`) [:math:`m`] - Suggested range: internal_radius >= 0.0
    :param external_radius: Initial external radius of the region (:math:`b_0`) [:math:`m`] - Suggested range: external_radius >= 0.0
    :param internal_pressure: Internal pressure applied on the inside of the sphere (:math:`p`) [:math:`kPa`] - Suggested range: internal_pressure >= 0.0
    :param external_pressure: External pressure on the outside of the sphere (:math:`p_0`) [:math:`kPa`] - Suggested range: external_pressure >= 0.0
    :param youngs_modulus: Young's modulus of the material (:math:`E`) [:math:`kPa`] - Suggested range: youngs_modulus >= 0.0
    :param poissons_ratio: Poisson's ratio of the material (:math:`\\nu`) [:math:`-`] - Suggested range: 0 <= poissons_ratio <= 0.5
    :param seed: Number of radii at which stresses and displacements are calculated (:math:``) [:math:`-`] (optional, default= 100)

    .. math::
        \\text{Elastic solutions}

        \\sigma_r = -p_0 - (p - p_0) \\frac{\\left( \\frac{b_0}{r} \\right)^3 - 1}{\\left( \\frac{b_0}{a_0} \\right)^3 - 1}

        \\sigma_{\\theta} = \\sigma_{\\phi} = -p_0 + (p - p_0) \\frac{\\frac{1}{2} \\left( \\frac{b_0}{r} \\right)^3 - 1}{\\left( \\frac{b_0}{a_0} \\right)^3 - 1}

        u = r - r_0 = \\frac{p - p_0}{E} \\frac{(1 - 2 \\nu) r + \\frac{(1 + \\nu) b_0^3}{2 r^2}}{\\left( \\frac{b_0}{a_0} \\right)^3 - 1}

        \\text{Yield criterion}

        \\sigma_1 - \\sigma_3 = 2 \\cdot S_u

        \\text{Internal pressure for yielding } (\\sigma_1 = \\sigma_{\\theta}, \\sigma_3 = \\sigma_r)

        p = p_{1y} = p_0 + \\frac{4 \\cdot S_u}{3} \\left[ 1 - \\left( \\frac{a_0}{b_0} \\right)^3 \\right]

        \\text{Displacement at internal and external boundaries}

        u |_{r=a_0} = \\frac{2 \\cdot S_u \\cdot a_0}{E} \\left[ \\frac{2 \\cdot (1 - 2 \\nu) \\cdot a_0^3 }{3 \\cdot b_0^3} + \\frac{1 + \\nu}{3} \\right]

        u |_{r=b_0} = \\frac{2 \\cdot S_u \\cdot (1 - \\nu) \\cdot a_0 }{E \\cdot b_0^2}

        \\text{Stresses and displacements in the elastic region after yielding}

        \\sigma_r = - \\frac{4 \\cdot S_u \\cdot c^3}{3 \\cdot b_0^3} \\left[ \\left( \\frac{b_0}{r} \\right)^3 - 1 \\right] - p_0

        \\sigma_{\\theta} = \\sigma_{\\phi} = \\frac{4 \\cdot S_u \\cdot c^3}{3 \\cdot b_0^3} \\left[ \\frac{1}{2} \\left( \\frac{b_0}{r} \\right)^3 + 1 \\right] - p_0

        u = \\frac{4 \\cdot S_u \\cdot c^3}{3 \\cdot E \\cdot b_0^3} \\left[ (1 - 2 \\nu) \\cdot r + \\frac{(1 + \\nu) b_0^3}{2 \\cdot r^2} \\right]

        \\text{Stresses in the plastic region}

        \\sigma_r = - 4 \\cdot S_u \\cdot \\ln \\left( \\frac{c}{r} \\right) - \\frac{4 \\cdot S_u}{3} \\left[ 1 - \\left( \\frac{c}{b_0} \\right)^3 \\right] - p_0

        \\sigma_{\\theta} = 2 \\cdot S_u - 4 \\cdot S_u \\cdot \\ln \\left( \\frac{c}{r} \\right) - \\frac{4 \\cdot S_u}{3} \\cdot \\left[ 1 - \\left( \\frac{c}{b_0} \\right)^3 \\right] - p_0

        \\text{The pressure required to generate a plastic radius is thus}

        p = 4 \\cdot S_u \\cdot \\ln \\left( \\frac{c}{a} \\right) + \\frac{4 \\cdot S_u}{3} \\left[ 1 - \\left( \\frac{c}{b_0} \\right)^3 \\right] + p_0

        \\text{Expansion of the boundary}

        \\left( \\frac{a}{a_0} \\right)^3 = 1 + \\frac{6 (1 - \\nu) S_u c^3}{E \\cdot a_0^3} - \\frac{4 (1 - 2 \\nu) S_u}{E} \\left[ 3 \\ln \\left( \\frac{c}{a_0} \\right) + 1 - \\left( \\frac{c}{b_0} \\right)^3 \\right]

    :returns: Dictionary with the following keys:

        - 'elastic radii [m]': Radii at which elastic stresses are calculated
        - 'elastic_radial_stress [kPa]': Radial stresses for purely elastic deformation (:math:`\\sigma_{r,elastic}`)  [:math:`kPa`]
        - 'elastic_tangential_stress [kPa]': Tangential stresses for purely elastic deformation (:math:`\\sigma_{\\theta,elastic}`)  [:math:`kPa`]
        - 'elastic_radial_displacement [m]': Radial displacement for purely elastic deformation (:math:`u_{elastic}`)  [:math:`m`]
        - 'yielding_pressure [kPa]': Internal pressure which initiates yield (:math:`p_{1y}`)  [:math:`kPa`]
        - 'plastic_radius [m]': Radius of the plastic zone for the given internal and external pressures (:math:`c`)  [:math:`m`]
        - 'elastoplastic_radii [m]': Radii at which elastoplastic stresses are calculated (equal amount of point inside and outside the plastic radius)
        - 'elastoplastic_radial_stress [kPa]': Radial stresses for Tresca soil (:math:`\\sigma_r`)  [:math:`kPa`]
        - 'elastoplastic_tangential_stress [kPa]': Tangential stresses for Tresca soil (:math:`\\sigma_{\\theta}`)  [:math:`kPa`]
        - 'expanded_radius [m]': Radius of the internal wall after expansion (:math:`a`)  [:math:`m`]

    Reference - Yu, H.-S., 2000. Cavity Expansion Methods in Geomechanics. Springer-Science+Business Media, B.V.

    """
    _radii = np.linspace(internal_radius, external_radius, seed)
    _elastic_radial_stress = -external_pressure - (internal_pressure - external_pressure) * \
                             ((((external_radius / _radii) ** 3) - 1) /
                              (((external_radius / internal_radius) ** 3) - 1))
    _elastic_tangential_stress = -external_pressure + (internal_pressure - external_pressure) * \
                                 ((0.5 * ((external_radius / _radii) ** 3) - 1) /
                                  (((external_radius / internal_radius) ** 3) - 1))
    _elastic_radial_displacement = ((internal_pressure - external_pressure) / youngs_modulus) * \
                                   (((1 - 2 * poissons_ratio) * _radii +
                                     (((1 + poissons_ratio) * (external_radius ** 3)) /
                                      (2 * _radii ** 2))) / (((external_radius / internal_radius) ** 3) - 1))
    _yielding_pressure = internal_pressure + (4 * undrained_shear_strength / 3) * \
                         (1 - ((internal_radius / external_radius) ** 3))

    def wall_expansion(c, a0, b0, su, E, nu):
        return a0 * ((
            1 +
            ((6 * (1 - nu) * su * (c ** 3)) / (E * (a0 ** 3))) -
            ((4 * (1 - 2 * nu) * su) / (E)) * (3 * np.log(c / a0) +
                                               1 -
                                               ((c / b0) ** 3))) ** (1 / 3))

    def optimisation_func(c, a0, b0, su, p, p0, E, nu):
        a = wall_expansion(c, a0, b0, su, E, nu)
        return 4 * su * np.log(c / a) + (4 * su / 3) * (1 - ((c / b0) ** 3)) + p0 - p

    _plastic_radius = None
    _external_radius = external_radius

    while _plastic_radius is None:
        try:
            _plastic_radius = brentq(
                f=optimisation_func,
                a=internal_radius,
                b=_external_radius,
                args=(internal_radius, external_radius,
                      undrained_shear_strength, internal_pressure, external_pressure,
                      youngs_modulus, poissons_ratio)
            )
        except Exception as err:
            _external_radius = 0.95 * _external_radius
            if _external_radius <= internal_radius:
                raise ValueError("A feasible plastic radius could not be found. The internal pressure might be too"
                                 "low to cause plastic deformation or too high"
                                 "to find a feasible plastic radius - %s" % (str(err)))

    _expanded_radius = wall_expansion(
        c=_plastic_radius,
        a0=internal_radius,
        b0=external_radius,
        su=undrained_shear_strength,
        E=youngs_modulus,
        nu=poissons_ratio
    )

    _elastoplastic_radii_elastic = np.linspace(_plastic_radius, external_radius, seed)
    _elastoplastic_radii_plastic = np.linspace(_expanded_radius, _plastic_radius, seed)
    _elastoplastic_radii = np.append(
        _elastoplastic_radii_plastic,
        _elastoplastic_radii_elastic)
    _elastoplastic_radial_stress_elastic = \
        -((4 * undrained_shear_strength * (_plastic_radius ** 3)) / (3 * (external_radius ** 3))) * \
        (((external_radius / _elastoplastic_radii_elastic) ** 3) - 1) - external_pressure
    _elastoplastic_radial_stress_plastic = \
        -4 * undrained_shear_strength * np.log(_plastic_radius / _elastoplastic_radii_plastic) - \
        (4 * undrained_shear_strength / 3) * (1 - ((_plastic_radius / external_radius) ** 3)) - \
        external_pressure
    _elastoplastic_tangential_stress_elastic = \
        -((4 * undrained_shear_strength * (_plastic_radius ** 3)) / (3 * (external_radius ** 3))) * \
        (0.5 * ((external_radius / _elastoplastic_radii_elastic) ** 3) + 1) - external_pressure
    _elastoplastic_tangential_stress_plastic = \
        2 * undrained_shear_strength - \
        4 * undrained_shear_strength * np.log(_plastic_radius / _elastoplastic_radii_plastic) - \
        (4 * undrained_shear_strength / 3) * (1 - ((_plastic_radius / external_radius) ** 3)) - \
        external_pressure
    _elastoplastic_radial_stress = np.append(
        _elastoplastic_radial_stress_plastic, _elastoplastic_radial_stress_elastic)
    _elastoplastic_tangential_stress = np.append(
        _elastoplastic_tangential_stress_plastic, _elastoplastic_tangential_stress_elastic)

    return {
        'elastic radii [m]': _radii,
        'elastic_radial_stress [kPa]': _elastic_radial_stress,
        'elastic_tangential_stress [kPa]': _elastic_tangential_stress,
        'elastic_radial_displacement [m]': _elastic_radial_displacement,
        'yielding_pressure [kPa]': _yielding_pressure,
        'plastic_radius [m]': _plastic_radius,
        'elastoplastic radii [m]': _elastoplastic_radii,
        'elastoplastic_radial_stress [kPa]': _elastoplastic_radial_stress,
        'elastoplastic_tangential_stress [kPa]': _elastoplastic_tangential_stress,
        'expanded_radius [m]': _expanded_radius,
    }


EXPANSION_CYLINDER_TRESCA = {
    'insitu_pressure': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'borehole_pressure': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'diameter': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'undrained_shear_strength': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'shear_modulus': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'poissons_ratio': {'type': 'float', 'min_value': 0.0, 'max_value': 0.5},
    'max_radius_multiplier': {'type': 'float', 'min_value': 1.0, 'max_value': None},
    'number_radii': {'type': 'int', 'min_value': None, 'max_value': None},
}

EXPANSION_CYLINDER_TRESCA_ERRORRETURN = {
    'yielding': False,
    'pressure expansion function': dict(),
    'yielding pressure [kPa]': np.nan,
    'radii [m]': np.nan,
    'radial stresses [kPa]': np.nan,
    'tangential stresses [kPa]': np.nan,
    'elastic wall expansion [m]': np.nan,
    'plastic wall expansion [m]': np.nan,
    'plastic radius [m]': np.nan,
}


@Validator(EXPANSION_CYLINDER_TRESCA, EXPANSION_CYLINDER_TRESCA_ERRORRETURN)
def expansion_cylinder_tresca(
        insitu_pressure, borehole_pressure, diameter, undrained_shear_strength, shear_modulus,
        poissons_ratio=0.5, max_radius_multiplier=10.0, number_radii=250, **kwargs):
    """
    Calculates the cavity expansion for a cylinder in Tresca material. The relation between borehole radius + plastic radius and pressure differential is calculated first.

    This relation is then used to evaluate where the imposed pressure lies, whether it causes any plasticity around the borehole and whether it does not cause overall borehole failure.

    The stresses for the given pressure are then calculated.

    :param insitu_pressure: Isotropic horizontal stress in the soil mass before borehole excavation (:math:`p_0`) [:math:`kPa`] - Suggested range: insitu_pressure >= 0.0
    :param borehole_pressure: Pressure on the borehole wall due to drilling fluids or concrete (:math:`p`) [:math:`kPa`] - Suggested range: borehole_pressure >= 0.0
    :param diameter: Borehole initial diameter (equations are formulated in terms of radius but diameter is more convenient as input) (:math:`2 \\cdot a_0`) [:math:`m`] - Suggested range: diameter >= 0.0
    :param undrained_shear_strength: Undrained shear strength of the material surrounding the borehole (:math:`S_u`) [:math:`kPa`] - Suggested range: undrained_shear_strength >= 0.0
    :param shear_modulus: Shear modulus of the material surrounding the borehole (:math:`G`) [:math:`kPa`] - Suggested range: shear_modulus >= 0.0
    :param poissons_ratio: Poissons ratio of the material surrounding the borehole (default for undrained material)) (:math:`\\nu`) [:math:`-`] - Suggested range: 0.0 <= poissons_ratio <= 0.5 (optional, default= 0.5)
    :param max_radius_multiplier: Multiplier on borehole radius for determining the maximum extent of the calculation (:math:``) [:math:`-`] - Suggested range: max_radius_multiplier >= 1.0 (optional, default= 10.0)
    :param number_radii: Number of radii considered for the calculation (:math:``) [:math:`-`] (optional, default= 250)

    .. math::
        \\text{Elastic properties}

        n = \\frac{4 \\cdot S_u \\cdot (1 - \\nu^2) }{E}

        E = 2 \\cdot G \\cdot (1 + \\nu)

        \\text{Plastic radius and wall radius expansion during plastic deformation}

        \\left( \\frac{c}{a} \\right)^2 = \\left( \\frac{a_0}{a} \\right)^2 + \\frac{1}{n} \\cdot \\left[ 1 - \\left( \\frac{a_0}{a} \\right)^2 \\right]

        \\frac{p - p_0}{2 \\cdot S_u} = \\frac{1}{2} + \\frac{1}{2} \\cdot \\ln \\left[ \\frac{G}{S_u} \\cdot \\left( 1 - \\left( \\frac{a_0}{a} \\right)^2 \\right) + \\left( \\frac{a_0}{a} \\right)^2 \\right]

        \\text{Elastic stresses and displacements can be calculated by taking the limit for the outer radius going to infinity}

        \\sigma_r = - p_0 - (p - p_0) \\cdot \\frac{\\frac{b_0^2}{r^2} - 1}{\\frac{b_0^2}{a_0^2} - 1}  \\implies \\sigma_r = -p_0 - (p - p_0) \\cdot \\frac{a_0^2}{r^2}

        \\sigma_{\\theta} = -p_0 + (p - p_0) \\cdot \\frac{\\frac{b_0^2}{r^2} + 1}{\\frac{b_0^2}{a_0^2} - 1} \\implies \\sigma_{\\theta} = -p_0 + (p - p_0) \\cdot \\frac{a_0^2}{r^2}

        u = \\frac{(1 + \\nu) \\cdot (p - p_0)}{E} \\cdot \\frac{a_0^2}{b_0^2 - a_0^2} \\cdot \\left[ (1 - 2 \\cdot \\nu) \\cdot r + \\frac{b_0^2}{r} \\right] \\implies u = \\frac{(1 + \\nu) \\cdot (p - p_0)}{E} \\cdot \\frac{a_0^2}{r}

        \\text{The plasticity critertion can be expressed as:}

        \\sigma_{r,r=a_0} - \\sigma_{theta,r=a_0} = 2 \\cdot S_u \\implies -(p - p_0)  = S_u

        \\text{The elastic stresses and displacements outside of the plastic zone can be written as:}

        \\sigma_r = -\\frac{S_u \\cdot c^2}{b_0^2} \\cdot \\left( \\frac{b_0^2}{r^2} - 1 \\right) - p_0 \\implies \\sigma_r = -\\frac{S_u \\cdot c^2}{r^2} - p_0

        \\sigma_{\\theta} = \\frac{S_u \\cdot c^2}{b_0^2} \\cdot \\left( \\frac{b_0^2}{r^2} + 1 \\right) - p_0 \\implies \\sigma_{\\theta} = \\frac{S_u \\cdot c^2}{r^2} - p_0

        u = \\frac{(1 + \\nu) \\cdot S_u \\cdot c^2}{E \\cdot b_0^2} \\cdot \\left[ (1 - 2 \\cdot \\nu) \\cdot r + \\frac{b_0^2}{r} \\right] \\implies u = \\frac{(1 + \\nu) \\cdot S_u \\cdot c^2}{E} \\cdot \\frac{1}{r}

        \\text{Stresses in the plastic zone:}

        \\sigma_r = -p_0 - S_u - 2 \\cdot S_u \\cdot \\ln \\left( \\frac{c}{r} \\right)

        \\sigma_{\\theta} = -p_0 + S_u - 2 \\cdot S_u \\cdot \\ln \\left( \\frac{c}{r} \\right)

    :returns: Dictionary with the following keys:

        - 'yielding': Boolean determining whether plastic deformation is taking place or not
        - 'pressure expansion function': Dictionary with the pressure-expansion relation with keys `expansion [m]` and `pressure difference [kPa]`
        - 'yielding pressure [kPa]': Borehole pressure at which yield occurs [:math:`kPa`]
        - 'radii [m]': Numpy array with radii used for the stress and displacement calculation [:math:`m`]
        - 'radial stresses [kPa]': Numpy array with the radial stresses around the borehole [:math:`kPa`]
        - 'tangential stresses [kPa]': Numpy array with the tangential stresses around the borehole [:math:`kPa`]
        - 'elastic wall expansion [m]': Borehole elastic wall expansion [:math:`m`]
        - 'plastic wall expansion [m]': Borehole plastic wall expansion (:math:`a - a_0`)  [:math:`m`]
        - 'plastic radius [m]': Radius of the plastic zone (:math:`c`)  [:math:`m`]

    Reference - Yu, H.-S., 2000. Cavity Expansion Methods in Geomechanics. Springer-Science+Business Media, B.V.

    """
    _yielding_pressure = insitu_pressure + undrained_shear_strength
    _youngs_modulus = 2 * shear_modulus * (1 + poissons_ratio)
    if abs(insitu_pressure - borehole_pressure) < undrained_shear_strength:
        _yielding = False
    else:
        _yielding = True
    _borehole_radius = 0.5 * diameter
    _radii = np.linspace(_borehole_radius, max_radius_multiplier * _borehole_radius, number_radii)
    _pressure_expansion_function = {
        'expansion [m]': _radii - _borehole_radius,
        'pressure difference [kPa]': undrained_shear_strength * (
                1 + np.log((shear_modulus/undrained_shear_strength) +
                           (1 - shear_modulus/undrained_shear_strength) * ((_borehole_radius / _radii) ** 2)))
    }

    def _elastoplastic_radialstress(r, c, su, p0):
        if r < c:
            # Plastic zone
            return -p0 - su - 2 * su * np.log(c / r)
        else:
            # Elastic zone
            return -su * ((c ** 2) / (r ** 2)) - p0

    def _elastoplastic_tangentialstress(r, c, su, p0):
        if r < c:
            # Plastic zone
            return -p0 + su - 2 * su * np.log(c / r)
        else:
            # Elastic zone
            return su * ((c ** 2) / (r ** 2)) - p0

    if (borehole_pressure - insitu_pressure) > _pressure_expansion_function['pressure difference [kPa]'].max():
        warnings.warn("Selected borehole pressure leads to excessive deformation. Stresses are not calculated.")
        _radial_stresses = np.array(list(map(lambda _x: np.nan, _radii)))
        _tangential_stresses = np.array(list(map(lambda _x: np.nan, _radii)))
        _elastic_wall_expansion = np.nan
        _plastic_wall_expansion = np.nan
        _plastic_radius = np.nan
    else:
        if _yielding:
            # Calculate the plastic radius
            _first_term = (_borehole_radius ** 2) * (1 - shear_modulus / undrained_shear_strength)
            _second_term = np.exp(((borehole_pressure - insitu_pressure) / undrained_shear_strength) - 1) - \
                           (shear_modulus / undrained_shear_strength)
            _expanded_radius = np.sqrt(_first_term / _second_term)
            _plastic_wall_expansion = _expanded_radius - _borehole_radius
            _n = 4 * (1 + (poissons_ratio ** 2)) * undrained_shear_strength / _youngs_modulus
            _plastic_radius = _expanded_radius * np.sqrt(
                ((_borehole_radius / _expanded_radius) ** 2) +
                (1 / _n) * (1 - ((_borehole_radius / _expanded_radius) ** 2)))

            _radial_stresses = np.array(
                list(map(lambda _r: _elastoplastic_radialstress(
                    r=_r, c=_plastic_radius, su=undrained_shear_strength, p0=insitu_pressure), _radii)))
            _tangential_stresses = np.array(
                list(map(lambda _r: _elastoplastic_tangentialstress(
                    r=_r, c=_plastic_radius, su=undrained_shear_strength, p0=insitu_pressure), _radii)))
        else:
            _radial_stresses = -insitu_pressure - (borehole_pressure - insitu_pressure) * \
                               ((_borehole_radius ** 2) / (_radii ** 2))
            _tangential_stresses = -insitu_pressure + (borehole_pressure - insitu_pressure) * \
                                   ((_borehole_radius ** 2) / (_radii ** 2))
            _plastic_wall_expansion = 0
            _plastic_radius = _borehole_radius
        _elastic_wall_expansion = ((1 + poissons_ratio) * (borehole_pressure - insitu_pressure)) / _youngs_modulus

    return {
        'yielding': _yielding,
        'pressure expansion function': _pressure_expansion_function,
        'yielding pressure [kPa]': _yielding_pressure,
        'radii [m]': _radii,
        'radial stresses [kPa]': _radial_stresses,
        'tangential stresses [kPa]': _tangential_stresses,
        'elastic wall expansion [m]': _elastic_wall_expansion,
        'plastic wall expansion [m]': _plastic_wall_expansion,
        'plastic radius [m]': _plastic_radius,
    }


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import unittest
import os

# 3rd party packages
import pandas as pd
import numpy as np

# Project imports
from groundhog.deepfoundations.axialcapacity import negativeskinfriction

TESTS_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


class Test_negativeskinfriction_pilegroup_zeevaertdebeer(unittest.TestCase):

    def setUp(self):
        K0 = 0.8
        DELTA = 25  # Interface friction angle, deg
        OD = 0.4  # Pile outer diameter, m
        INFLUENCE_ZONE = 3 / 0.4  # Multiplier on pile outer diameter
        GAMMA_EFF = 8  # Effective unit weight of soil, kN/m3
        SURCHARGE = 51  # kPa
        MAX_DEPTH = 4.5
        self.depth = np.linspace(0, MAX_DEPTH, 250)
        self.K0 = np.ones(250) * K0
        self.delta = np.ones(250) * DELTA
        self.diameter = OD
        self.diameter_influence = OD * INFLUENCE_ZONE
        self.gamma_eff = np.ones(250) * GAMMA_EFF
        self.p0 = SURCHARGE

    def test_negativeskinfriction_pilegroup_zeevaertdebeer(self):
        result = negativeskinfriction.negativeskinfriction_pilegroup_zeevaertdebeer(
            depths=self.depth,
            effective_unit_weights=self.gamma_eff,
            lateral_earth_pressure_coefficients=self.K0,
            interface_friction_angles=self.delta,
            surcharge=self.p0,
            diameter=self.diameter,
            diameter_influence=self.diameter_influence,
            fail_silently=False
        )
        self.assertAlmostEqual(
            result['negative_skin_friction [kN]'], 145.56, 2
        )
        self.assertAlmostEqual(
            result['negative_skin_friction_group [kN]'], 127.19, 2
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Django and native Python packages

# 3rd party packages
import pandas as pd
# Project imports


def load_ags(path, groupname, **kwargs):
    """
    Generates a Pandas dataframe from an .ags file for a given group name.
    The function is based on the AGS 4.0 standard for geotechnical data transfer

    :param path: Path to the file
    :param groupname: AGS 4.0 group name
    :return: A pandas dataframe with the AGS data
    """

    # Read the raw AGS data
    with open(path, 'rb') as f:
        _lines = list(map(lambda _x: _x.decode('utf-8', 'backslashreplace'), f.readlines()))
        ags_raw = pd.DataFrame(_lines, columns=["AGS lines"])

    # Define where the data for the selected group starts
    _start_index = ags_raw[ags_raw['AGS lines'].str.startswith(r'"GROUP","%s"' % groupname)].index[0]

    # Define where the data for the selected group ends
    empty_rows = ags_raw[(ags_raw['AGS lines'] == "\r\n") | (ags_raw['AGS lines'] == "\n")]
    _end_index = empty_rows[empty_rows.index > _start_index].index[0]

    # Read only the data for the group
    _group_data = pd.read_csv(path, skiprows=_start_index + 1, nrows=_end_index - _start_index - 2, **kwargs)

    # Rename the headers
    new_headers = []
    datatypes = dict()
    for i, original_header in enumerate(_group_data.columns):
        if str(_group_data.loc[0, original_header]) == 'nan':
            new_name = "%s" % (original_header)
        else:
            new_name = "%s [%s]" % (original_header, _group_data.loc[0, original_header])
        new_headers.append(new_name)
        if "DP" in _group_data.loc[1, original_header]:
            datatypes[new_name] = 'float'
    _group_data.columns = new_headers

    _group_data = _group_data[2:].reset_index(drop=True).astype(datatypes)

    return _group_data


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages

# 3rd party packages
import numpy as np

# Project imports
from groundhog.general.validation import Validator


PILE_SETTLEMENT_CURVES = {
    'diameter': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'shaft_resistance': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'base_resistance': {'type': 'float', 'min_value': 0.0, 'max_value': None},
    'pile_type': {'type': 'string', 'options': ('driven', 'CFA', 'bored'), 'regex': None},
}

PILE_SETTLEMENT_CURVES_ERRORRETURN = {
    'shaft_normalised': None,
    'shaft_denormalised': None,
    'base_normalised': None,
    'base_denormalised': None,
    'total': None
}


@Validator(PILE_SETTLEMENT_CURVES, PILE_SETTLEMENT_CURVES_ERRORRETURN)
def pile_settlement_curves(
        diameter, shaft_resistance, base_resistance, pile_type,
        **kwargs):
    """
    Calculates the pile settlement curve for pile shaft and pile base from empirical trends established based on axial pile load tests.
    These curves take into account the pile type (driven, bored or CFA) and the fact that shaft resistance mobilisation distance does not appear to be proportional to the pile diameter, while the base resistance mobilisation distance is.
    The empirical curves are approximated by a mathematical relation. The function returns both the normalised and denormalised curves. Not that the function returns the pile settlement in m which the empirical shaft mobilisation curves mention mm.

    Finally, the overall curve is calculated by summing both curves

    :param diameter: Pile diameter (:math:`D`) [:math:`m`] - Suggested range: diameter >= 0.0
    :param shaft_resistance: Shaft resistance used to denormalise the curve (:math:`F_s`) [:math:`kN`] - Suggested range: shaft_resistance >= 0.0
    :param base_resistance: Base resistance used to denormalise the curve (:math:`F_b`) [:math:`kN`] - Suggested range: base_resistance >= 0.0
    :param pile_type: Pile type - Options: ('driven', 'CFA', 'bored')

    .. math::
        F_{mob} = a + \\frac{b-a}{1 + \\left( \\frac{\\delta_{pile} \\ \\text{or} \\ \\delta_{pile}/D}{c} \\right)^d}

    :returns: Dictionary with the following keys:

        - 'shaft_normalised': Normalised shaft mobilisation curve as a dictionary with keys 'w [m]' and 'Fs/Fsmax [-]'
        - 'shaft_denormalised': Shaft mobilisation curve as a dictionary with keys 'w [m]' and 'Fs [kN]'
        - 'base_normalised': Normalised base mobilisation curve as a dictionary with keys 'w/D [-]' and 'Fb/Fbmax [-]'
        - 'base_denormalised': Base mobilisation curve as a dictionary with keys 'w [m]' and 'Fb [kN]'
        - 'total': Pile tip settlement vs total load with keys 'w [m]' and 'F [kN]'

    .. figure:: images/pile_settlement_curves_1.png
        :figwidth: 500.0
        :width: 450.0
        :align: center

        Normalised shaft and base resistance mobilisation curves

    Reference - Syllabus geotechnics

    """

    coeff = {
        'Toe 1': {'a': 145.4, 'b': 0.58, 'c': 3.21, 'd': 0.70},
        'Toe 2': {'a': 137.57, 'b': 0.06, 'c': 5.63, 'd': 0.78},
        'Toe 3': {'a': 139.37, 'b': 0.0, 'c': 8.01, 'd': 1.0},
        'Shaft 1+2': {'a': 134.41, 'b': 3.49, 'c': 2.8, 'd': 0.82},
        'Shaft 3': {'a': 118.64, 'b': 2.8, 'c': 5.91, 'd': 1.23}
    }

    displ = np.linspace(0.0, 25.0, 250)

    # Driven
    fmob_toe_1 = np.minimum(coeff['Toe 1']['a'] + ((coeff['Toe 1']['b'] - coeff['Toe 1']['a']) /
                                                   (1 + ((displ / coeff['Toe 1']['c']) ** coeff['Toe 1']['d']))),
                            100.0 * np.ones(len(displ)))
    # CFA
    fmob_toe_2 = np.minimum(coeff['Toe 2']['a'] + ((coeff['Toe 2']['b'] - coeff['Toe 2']['a']) /
                                                   (1 + ((displ / coeff['Toe 2']['c']) ** coeff['Toe 2']['d']))),
                            100.0 * np.ones(len(displ)))
    # Bored
    fmob_toe_3 = np.minimum(coeff['Toe 3']['a'] + ((coeff['Toe 3']['b'] - coeff['Toe 3']['a']) /
                                                   (1 + ((displ / coeff['Toe 3']['c']) ** coeff['Toe 3']['d']))),
                            100.0 * np.ones(len(displ)))
    # Driven + CFA
    fmob_shaft_12 = np.minimum(coeff['Shaft 1+2']['a'] + ((coeff['Shaft 1+2']['b'] - coeff['Shaft 1+2']['a']) /
                                                          (1 + ((displ / coeff['Shaft 1+2']['c']) ** coeff['Shaft 1+2'][
                                                              'd']))),
                               100.0 * np.ones(len(displ)))
    # Bored
    fmob_shaft_3 = np.minimum(coeff['Shaft 3']['a'] + ((coeff['Shaft 3']['b'] - coeff['Shaft 3']['a']) /
                                                       (1 + ((displ / coeff['Shaft 3']['c']) ** coeff['Shaft 3'][
                                                           'd']))),
                              100.0 * np.ones(len(displ)))


    if pile_type == 'driven':
        fmob_toe = fmob_toe_1
        fmob_shaft = fmob_shaft_12
    elif pile_type == 'CFA':
        fmob_toe = fmob_toe_2
        fmob_shaft = fmob_shaft_12
    elif pile_type == 'bored':
        fmob_toe = fmob_toe_3
        fmob_shaft = fmob_shaft_3
    else:
        raise ValueError("Pile type not recognised")

    _shaft_normalised = {
        'w [m]': displ * 0.001,
        'Fs/Fsmax [-]': 0.01 * fmob_shaft
    }
    _shaft_denormalised = {
        'w [m]': displ * 0.001,
        'Fs [kN]': 0.01 * fmob_shaft * shaft_resistance
    }
    _base_normalised = {
        'w/D [-]': 0.01 * displ,
        'Fb/Fbmax [-]': 0.01 * fmob_toe
    }
    _base_denormalised = {
        'w [m]': 0.01 * displ * diameter,
        'Fb [kN]': 0.01 * fmob_toe * base_resistance
    }

    _w_max = max(_base_denormalised['w [m]'].max(), _shaft_denormalised['w [m]'].max())
    _w_total = np.linspace(0, _w_max, 1000)
    _fs_interp = np.interp(_w_total, _shaft_denormalised['w [m]'], _shaft_denormalised['Fs [kN]'])
    _fb_interp = np.interp(_w_total, _base_denormalised['w [m]'], _base_denormalised['Fb [kN]'])
    _f_total = _fs_interp + _fb_interp

    return {
        'shaft_normalised': _shaft_normalised,
        'shaft_denormalised': _shaft_denormalised,
        'base_normalised': _base_normalised,
        'base_denormalised': _base_denormalised,
        'total': {
            'w [m]': _w_total,
            'F [kN]': _f_total
        }
    }


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import unittest
import os

# 3rd party packages
import pandas as pd
import numpy as np

# Project imports
from groundhog.deepfoundations.boreholestability import cavityexpansion

TESTS_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


class Test_cavityexpansionfunctions(unittest.TestCase):

    def test_stress_elastic_isotropic(self):
        result = cavityexpansion.stress_cylinder_elastic_isotropic(
            radius=1, internal_pressure=0, farfield_pressure=100, borehole_radius=1,
            fail_silently=False
        )
        self.assertEqual(result['radial stress [kPa]'], 0)
        self.assertEqual(result['tangential stress [kPa]'], 200)
        result = cavityexpansion.stress_cylinder_elastic_isotropic(
            radius=1e7, internal_pressure=0, farfield_pressure=100, borehole_radius=1,
            fail_silently=False
        )
        self.assertAlmostEqual(result['radial stress [kPa]'], 100, 3)
        self.assertAlmostEqual(result['tangential stress [kPa]'], 100, 3)
        result = cavityexpansion.stress_cylinder_elastic_isotropic(
            radius=np.linspace(1, 1e7, 100), internal_pressure=0, farfield_pressure=100, borehole_radius=1,
            fail_silently=False
        )
        self.assertEqual(result['radial stress [kPa]'][0], 0)
        self.assertEqual(result['tangential stress [kPa]'][0], 200)
        self.assertAlmostEqual(result['radial stress [kPa]'][-1], 100, 3)
        self.assertAlmostEqual(result['tangential stress [kPa]'][-1], 100, 3)

    def test_expansion_tresca_thicksphere(self):
        result = cavityexpansion.expansion_tresca_thicksphere(
            undrained_shear_strength=10,
            internal_radius=1,
            external_radius=10,
            internal_pressure=150,
            external_pressure=100,
            youngs_modulus=1000,
            poissons_ratio=0.495,
            fail_silently=False)
        self.assertAlmostEqual(result['expanded_radius [m]'], 1.250, 3)

    def test_expansion_cylinder_tresca(self):
        elastic_result = cavityexpansion.expansion_cylinder_tresca(
            insitu_pressure=100,
            borehole_pressure=105,
            diameter=0.4,
            undrained_shear_strength=10,
            shear_modulus=1000)
        self.assertFalse(elastic_result['yielding'])
        self.assertAlmostEqual(elastic_result['elastic wall expansion [m]'], 0.0025, 4)
        self.assertEqual(elastic_result['plastic wall expansion [m]'], 0)
        self.assertEqual(elastic_result['plastic radius [m]'], 0.2)
        plastic_result = cavityexpansion.expansion_cylinder_tresca(
            insitu_pressure=100,
            borehole_pressure=120,
            diameter=0.4,
            undrained_shear_strength=10,
            shear_modulus=1000)
        self.assertTrue(plastic_result['yielding'])
        self.assertAlmostEqual(plastic_result['elastic wall expansion [m]'], 0.01, 3)
        self.assertAlmostEqual(plastic_result['plastic wall expansion [m]'], 0.0018, 4)
        collapsed_result = cavityexpansion.expansion_cylinder_tresca(
            insitu_pressure=100,
            borehole_pressure=250,
            diameter=0.4,
            undrained_shear_strength=10,
            shear_modulus=1000)
        self.assertTrue(collapsed_result['yielding'])
        self.assertTrue(np.math.isnan(collapsed_result['elastic wall expansion [m]']))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import unittest
import os

# 3rd party packages
import pandas as pd
import numpy as np

# Project imports
from groundhog.general import soilprofile as sp


class Test_SoilProfile(unittest.TestCase):

    def setUp(self):
        self.profile = sp.SoilProfile(
            {
                'Depth from [m]': [0, 1, 5, 10],
                'Depth to [m]': [1, 5, 10, 20],
                'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND']
            }
        )

    def test_wrong_layering(self):
        self.assertRaises(IOError, sp.SoilProfile, ({
            'Depth from [m]': [0, 1, 4, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND']
        }))

    def test_wrong_depthkey(self):
        self.assertRaises(IOError, sp.SoilProfile, ({
            'z from [m]': [0, 1, 4, 10],
            'z to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND']
        }))

    def test_wrong_linearvariation(self):
        # Test incomplete linear parameter variation
        self.assertRaises(IOError, sp.SoilProfile, ({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'qc from [MPa]': [1, 3, 10, 40]
        }))
        # Check that complete parameter variation passes
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [1, 3, 10, 40]
        })

    def test_profile_creation(self):
        self.assertEqual(self.profile.max_depth, 20)
        self.assertEqual(self.profile.min_depth, 0)
        self.assertEqual(self.profile.layer_transitions()[0], 1)
        self.assertEqual(self.profile.layer_transitions(include_top=True)[0], 0)

    def test_soilparameter_retrieval(self):
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'Dr [%]': [40, 60, np.nan, 80],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [1, 3, 10, 40]
        })
        self.assertIn('qc [MPa]',profile.soil_parameters())
        self.assertIn('Soil type',profile.soil_parameters())
        self.assertIn('qc from [MPa]',profile.soil_parameters(condense_linear=False))
        self.assertIn('qc [MPa]', profile.numerical_soil_parameters())
        self.assertNotIn('Soil type', profile.numerical_soil_parameters())
        self.assertIn('qc from [MPa]', profile.numerical_soil_parameters(condense_linear=False))
        self.assertNotIn('Soil type', profile.numerical_soil_parameters(condense_linear=False))
        self.assertNotIn('qc from [MPa]', profile.string_soil_parameters())
        self.assertIn('Soil type', profile.string_soil_parameters())
        self.assertTrue(profile.check_linear_variation('qc [MPa]'))
        self.assertFalse(profile.check_linear_variation('Dr [%]'))

    def test_transition_insert(self):
        self.profile.insert_layer_transition(depth=2.5)
        self.assertEqual(self.profile.loc[1, "Depth to [m]"], 2.5)
        self.profile.insert_layer_transition(depth=2)
        self.assertEqual(self.profile.loc[1, "Depth to [m]"], 2)

    def test_sign_conversion(self):
        self.profile.convert_depth_sign()
        self.assertEqual(self.profile.loc[1, 'Depth to [m]'], -5)
        self.profile.convert_depth_sign()
        self.assertEqual(self.profile.loc[1, 'Depth to [m]'], 5)

    def test_depth_shift(self):
        self.profile.shift_depths(offset=3)
        self.assertEqual(self.profile.loc[1, 'Depth to [m]'], 8)
        self.profile.shift_depths(offset=-3)
        self.assertEqual(self.profile.loc[1, 'Depth to [m]'], 5)

    def test_soilparameter_series(self):
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'Dr [%]': [40, 60, np.nan, 80],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [2, 3, 20, 50]
        })
        self.assertEqual(
            profile.soilparameter_series("qc [MPa]")[1][2], 3)
        self.assertEqual(
            profile.soilparameter_series("Dr [%]")[1][3], 60)

    def test_parameter_mapping(self):
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'Relative density': ['Loose', 'Medium dense', None, 'Dense'],
            'Dr [%]': [40, 60, np.nan, 80],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [2, 3, 20, 50]
        })
        mapped_df = profile.map_soilprofile(np.linspace(0, 20, 21))
        self.assertEqual(mapped_df.loc[3, 'Soil type'], 'SILT')
        self.assertRaises(ValueError, profile.map_soilprofile, (np.linspace(-1, 20, 22),))
        self.assertEqual(mapped_df.loc[15, 'qc [MPa]'], 45)

    def test_profile_selection_constant(self):
        self.profile.selection_soilparameter(
            parameter="Su [kPa]",
            depths=[6, 7, 8, 9],
            values=[10, 11, 12, 13]
        )
        self.assertEqual(self.profile.loc[2, "Su [kPa]"], 11.5)
        self.profile.selection_soilparameter(
            parameter="Dr [%]",
            depths=[6, 7, 8, 9],
            values=[10, 11, 12, 13],
            rule="min"
        )
        self.assertEqual(self.profile.loc[2, "Dr [%]"], 10)
        self.profile.selection_soilparameter(
            parameter="gamma [kN/m3]",
            depths=[6, 7, 8, 9],
            values=[10, 11, 12, 13],
            rule="max"
        )
        self.assertEqual(self.profile.loc[2, "gamma [kN/m3]"], 13)

    def test_profile_selection_linear(self):
        self.profile.selection_soilparameter(
            parameter="Su [kPa]",
            depths=[6, 7, 8, 9],
            values=[10, 11, 12, 13],
            linearvariation=True
        )
        self.assertAlmostEqual(self.profile.loc[2, "Su from [kPa]"], 9, 4)
        self.assertAlmostEqual(self.profile.loc[2, "Su to [kPa]"], 14, 4)
        # TODO: Check linear variation with min and max rule

    def test_merge_layers_top(self):
        self.profile.merge_layers(layer_ids=(1, 2))
        self.assertEqual(self.profile.loc[1, "Depth from [m]"], 1)
        self.assertEqual(self.profile.loc[1, "Depth to [m]"], 10)
        self.assertEqual(self.profile.loc[1, "Soil type"], "SILT")

    def test_merge_layers_bottom(self):
        self.profile.merge_layers(layer_ids=(1, 2), keep='bottom')
        self.assertEqual(self.profile.loc[1, "Depth from [m]"], 1)
        self.assertEqual(self.profile.loc[1, "Depth to [m]"], 10)
        self.assertEqual(self.profile.loc[1, "Soil type"], "CLAY")

    def test_remove_parameter(self):
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'Relative density': ['Loose', 'Medium dense', None, 'Dense'],
            'Dr [%]': [40, 60, np.nan, 80],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [2, 3, 20, 50]
        })
        profile.remove_parameter('qc [MPa]')
        def test_func():
            return profile.loc[0, "qc from [MPa]"]
        self.assertRaises(KeyError, test_func)
        profile.remove_parameter('Dr [%]')
        def test_func_2():
            return profile.loc[0, "Dr [%]"]

        self.assertRaises(KeyError, test_func_2)

    def test_cut_profile(self):
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'Relative density': ['Loose', 'Medium dense', None, 'Dense'],
            'Dr [%]': [40, 60, np.nan, 80],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [2, 4, 20, 50]
        })
        pf = profile.cut_profile(top_depth=1.5, bottom_depth=19)
        self.assertEqual(pf.min_depth, 1.5)
        self.assertEqual(pf.max_depth, 19)
        self.assertEqual(pf.loc[0, "qc from [MPa]"], 3.125)
        self.assertEqual(pf["qc to [MPa]"].iloc[-1], 49)

    def test_depth_integration(self):
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'Relative density': ['Loose', 'Medium dense', None, 'Dense'],
            'Unit weight [kN/m3]': [9, 8, 7, 10],
            'Dr [%]': [40, 60, np.nan, 80],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [2, 4, 20, 50]
        })
        profile.depth_integration(
            parameter='Unit weight [kN/m3]',
            outputparameter='Total vertical stress [kPa]'
        )
        self.assertEqual(profile['Total vertical stress from [kPa]'].iloc[0], 0)
        self.assertEqual(profile['Total vertical stress to [kPa]'].iloc[0], 9)
        self.assertEqual(profile['Total vertical stress to [kPa]'].iloc[1], 41)

    def test_convert_to_constant(self):
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'Relative density': ['Loose', 'Medium dense', None, 'Dense'],
            'Total unit weight from [kN/m3]': [19, 18, 17, 20],
            'Total unit weight to [kN/m3]': [19, 18, 19, 20],
            'Dr [%]': [40, 60, np.nan, 80],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [2, 4, 20, 50]
        })
        profile.convert_to_constant("qc [MPa]")
        profile.convert_to_constant("Total unit weight [kN/m3]", rule='min')
        self.assertEqual(profile['qc [MPa]'].iloc[0], 1.5)
        self.assertEqual(profile['Total unit weight [kN/m3]'].iloc[2], 17)

    def test_calculate_overburden(self):
        profile = sp.SoilProfile({
            'Depth from [m]': [0, 1, 5, 10],
            'Depth to [m]': [1, 5, 10, 20],
            'Soil type': ['SAND', 'SILT', 'CLAY', 'SAND'],
            'Relative density': ['Loose', 'Medium dense', None, 'Dense'],
            'Total unit weight [kN/m3]': [19, 18, 17, 20],
            'Dr [%]': [40, 60, np.nan, 80],
            'qc from [MPa]': [1, 3, 10, 40],
            'qc to [MPa]': [2, 4, 20, 50]
        })
        profile.calculate_overburden(waterlevel=4)
        self.assertEqual(profile['Total vertical stress to [kPa]'].iloc[-1], 376)
        self.assertEqual(profile['Effective vertical stress to [kPa]'].iloc[-1], 216)
        self.assertEqual(profile['Hydrostatic pressure to [kPa]'].iloc[-1], 160)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import unittest
import os

# 3rd party packages
import pandas as pd
import numpy as np

# Project imports
from groundhog.deepfoundations.axialcapacity import koppejan

TESTS_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


class Test_pilecalckoppejan(unittest.TestCase):

    def setUp(self):
        self.cpt_data = pd.read_excel(os.path.join(TESTS_DATA_DIR, 'cpt_koppejan.xlsx'))
        self.layering = pd.DataFrame({
            'z from [m]': [0, 3, 7, 14, 18.2, 21.3],
            'z to [m]': [3, 7, 14, 18.2, 21.3, 22],
            'Total unit weight [kN/m3]': [17, 20, 18, 20, 20, 20]
        })

    def test_mapping(self):
        """
        Tests the mapping of soil parameters to the grid
        """
        self.koppejan = koppejan.KoppejanCalculation(
            depth=self.cpt_data['z [m]'],
            qc=self.cpt_data['qc [MPa]'],
            diameter=0.4,
            penetration=16.5
        )
        self.koppejan.set_layer_properties(
            layer_data=self.layering)
        self.assertEqual(
            np.interp(
                5,
                self.koppejan.data['z [m]'],
                self.koppejan.data['Total unit weight [kN/m3]']), 20
        )
        self.assertEqual(
            np.interp(
                21.5,
                self.koppejan.data['z [m]'],
                self.koppejan.data['qclim [MPa]']), 12
        )

    def test_shaft_capacity(self):
        """
        Test the shaft capacity calculation according to Koppejan
        """
        self.test_mapping()
        self.koppejan.calculate_side_friction(alpha_s=0.01)
        self.assertAlmostEqual(self.koppejan.Frs, 1132, 0)

    def test_base_capacity(self):
        """
        Test the base capacity calculation according to Koppejan
        """
        self.test_mapping()
        self.koppejan.calculate_base_resistance(alpha_p=1)
        self.assertAlmostEqual(self.koppejan.qcII, 16.13, 2)
        self.assertAlmostEqual(self.koppejan.qcIII, 8.02, 2)
        self.assertAlmostEqual(self.koppejan.qcI, 12.51, 2)
        self.assertAlmostEqual(self.koppejan.Frb, 1404, 0)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import unittest
import os

# 3rd party packages
import numpy as np

# Project imports
from groundhog.siteinvestigation.insitutests import pcpt_correlations


TESTS_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


class Test_behaviourindex_pcpt_robertsonwride(unittest.TestCase):

    def test_behaviourindex_pcpt_robertsonwride(self):
        result = pcpt_correlations.behaviourindex_pcpt_robertsonwride(
            qt=40,
            fs=0.1,
            sigma_vo=150,
            sigma_vo_eff=50
        )
        self.assertAlmostEqual(result['Ic [-]'], 1.005, 3)
        result = pcpt_correlations.behaviourindex_pcpt_robertsonwride(
            qt=1,
            fs=0.1,
            sigma_vo=150,
            sigma_vo_eff=50
        )
        self.assertAlmostEqual(result['Ic [-]'], 3.203, 3)


class Test_gmax_sand_rixstokoe(unittest.TestCase):

    def test_gmax_sand_rixstokoe(self):
        result = pcpt_correlations.gmax_sand_rixstokoe(
            qc=15,
            sigma_vo_eff=100
        )
        self.assertAlmostEqual(result['Gmax [kPa]'], 101689.2, 1)


class Test_gmax_clay_maynerix(unittest.TestCase):

    def test_gmax_clay_maynerix(self):
        result = pcpt_correlations.gmax_clay_maynerix(
            qc=1
        )
        self.assertAlmostEqual(result['Gmax [kPa]'], 28121.9, 1)


class Test_relativedensity_ncsand_baldi(unittest.TestCase):

    def test_relativedensity_ncsand_baldi(self):
        result = pcpt_correlations.relativedensity_ncsand_baldi(
            qc=20,
            sigma_vo_eff=200
        )
        self.assertAlmostEqual(result['Dr [-]'], 0.8, 1)


class Test_relativedensity_ocsand_baldi(unittest.TestCase):

    def test_relativedensity_ocsand_baldi(self):
        result = pcpt_correlations.relativedensity_ocsand_baldi(
            qc=20,
            sigma_vo_eff=200,
            k0=1
        )
        self.assertAlmostEqual(result['Dr [-]'], 0.686, 3)


class Test_relativedensity_sand_jamiolkowski(unittest.TestCase):

    def test_relativedensity_sand_jamiolkowski(self):
        result = pcpt_correlations.relativedensity_sand_jamiolkowski(
            qc=20,
            sigma_vo_eff=200,
            k0=1
        )
        self.assertAlmostEqual(result['Dr dry [-]'], 0.60, 2)
        self.assertAlmostEqual(result['Dr sat [-]'], 0.68, 2)


class Test_friction_angle_kleven(unittest.TestCase):
    def test_values(self):
        self.assertEqual(
            pcpt_correlations.frictionangle_overburden_kleven(
                10.0, 100.0, Ko=1.0, max_friction_angle=50.0)['phi [deg]'], 47.497)
        self.assertEqual(
            pcpt_correlations.frictionangle_overburden_kleven(
                10.0, 100.0, Ko=1.0, )['phi [deg]'], 45.0)

    def test_ranges(self):
        self.assertRaises(
            ValueError,
            pcpt_correlations.frictionangle_overburden_kleven, 1.0, 100.0, Ko=0.6, fail_silently=False)


class Test_ocr_cpt_lunne(unittest.TestCase):

    def test_values(self):
        result = pcpt_correlations.ocr_cpt_lunne(
                Qt=10)
        self.assertAlmostEqual(result['OCR_Qt_LE [-]'], 2.28, 2)
        self.assertAlmostEqual(result['OCR_Qt_BE [-]'], 3.01, 2)
        self.assertAlmostEqual(result['OCR_Qt_HE [-]'], 4.68, 2)
        self.assertTrue(np.math.isnan(result['OCR_Bq_LE [-]']))

        result = pcpt_correlations.ocr_cpt_lunne(
            Qt=10, Bq=0.6)
        self.assertAlmostEqual(result['OCR_Qt_LE [-]'], 2.28, 2)
        self.assertAlmostEqual(result['OCR_Qt_BE [-]'], 3.01, 2)
        self.assertAlmostEqual(result['OCR_Qt_HE [-]'], 4.68, 2)
        self.assertAlmostEqual(result['OCR_Bq_LE [-]'], 1.41, 2)
        self.assertAlmostEqual(result['OCR_Bq_BE [-]'], 2.07, 2)
        self.assertAlmostEqual(result['OCR_Bq_HE [-]'], 3.09, 2)


class Test_sensitivity_frictionratio_lunne(unittest.TestCase):

    def test_values(self):
        result = pcpt_correlations.sensitivity_frictionratio_lunne(Rf=1)
        self.assertAlmostEqual(result['St LE [-]'], 5.71, 2)
        self.assertAlmostEqual(result['St BE [-]'], 7.47, 2)
        self.assertAlmostEqual(result['St HE [-]'], 9.63, 2)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import os
import warnings
from copy import deepcopy
import json
import re

# 3rd party packages
import pandas as pd
from plotly import tools, subplots
import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS
from plotly.offline import iplot
try:
    import pydov
    from owslib.fes import PropertyIsEqualTo
    from pydov.search.sondering import SonderingSearch
except:
    warnings.warn('pydov package not available. Import of CPT data from Databank Ondergrond Vlaanderen is not possible.'
                  'Install pydov to enable this functionality')

# Project imports
from groundhog.general.plotting import plot_with_log
from groundhog.general.parameter_mapping import map_depth_properties, merge_two_dicts, reverse_dict
from groundhog.siteinvestigation.insitutests.pcpt_correlations import *
from groundhog.general.soilprofile import SoilProfile


DEFAULT_CONE_PROPERTIES = SoilProfile({
    'Depth from [m]': [0, ],
    'Depth to [m]': [20, ],
    'area ratio [-]': [0.8, ],
    'Cone type': ['U', ],
    'Cone base area [cm2]': [10, ],
    'Cone sleeve_area [cm2]': [150, ]
})


class PCPTProcessing(object):
    """
    The PCPTProcessing class implements methods for reading, processing and presentation of PCPT data.
    Common correlations are also encoded.
    """

    def __init__(self, title, easting=np.nan, northing=np.nan, elevation=np.nan, srid=4326, waterunitweight=10.25):
        """
        Initialises a PCPTProcessing object based on a title. Optionally, a geographical position can be defined.

        An empty dataframe (`.data`) is created for storing the PCPT data

        :param title: Title for the PCPT test
        :param easting: Easting for test location (optional, default=NaN)
        :param northing: Northing for test location (optional, default=NaN)
        :param elevation: Elevation for test location (optional, default=NaN)
        :param srid: Projection ID for the location (optional, default=4325, see https://epsg.io/4326)
        :param waterunitweight: Unit weight of water used for effective stress calculations (default=10.25kN/m3 for seawater)

        """
        self.title = title
        self.easting = easting
        self.northing = northing
        self.elevation = elevation
        self.srid = srid
        self.data = pd.DataFrame()
        self.downhole_corrected = False
        self.waterunitweight=waterunitweight

    # region Utility functions

    def rename_columns(self, z_key=None, qc_key=None, fs_key=None, u2_key=None, push_key=None):
        if z_key is None:
            z_key = "z [m]"

        if qc_key is None:
            qc_key = "qc [MPa]"

        if fs_key is None:
            fs_key = "fs [MPa]"

        if u2_key is None:
            u2_key = "u2 [MPa]"

        if push_key is None:
            push_key = "Push"

        self.data.rename(columns={z_key: 'z [m]',
                                  qc_key: 'qc [MPa]',
                                  fs_key: 'fs [MPa]',
                                  u2_key: 'u2 [MPa]',
                                  push_key: 'Push'}, inplace=True)

    def convert_columns(self, qc_multiplier=1, fs_multiplier=1, u2_multiplier=1):
        self.data['qc [MPa]'] = self.data['qc [MPa]'] * qc_multiplier
        self.data['fs [MPa]'] = self.data['fs [MPa]'] * fs_multiplier
        self.data['u2 [MPa]'] = self.data['u2 [MPa]'] * u2_multiplier

    def add_zerodepth_row(self):
        self.data.loc[self.data.__len__(), "z [m]"] = 0
        self.data.sort_values("z [m]", inplace=True)
        self.data.reset_index(drop=True, inplace=True)

    def dropna_rows(self):
        self.data.dropna(subset=('z [m]', 'qc [MPa]', 'fs [MPa]', 'u2 [MPa]'), how='all', inplace=True)

    # endregion

    # region Data loading

    def load_excel(self, path, z_key=None, qc_key=None, fs_key=None, u2_key=None, push_key=None,
                   qc_multiplier=1, fs_multiplier=1, u2_multiplier=1, add_zero_row=True, **kwargs):
        """
        Loads PCPT data from an Excel file. Specific column keys have to be provided for z, qc, fs and u2.
        If column keys are not specified, the following keys are used:

            - 'z [m]' for depth below mudline
            - 'qc [MPa]' for cone tip resistance
            - 'fs [MPa]' for sleeve friction
            - 'u2 [MPa]' for pore pressure at cone shoulder

        Note that cone tip resistance, sleeve friction and pore pressure at the shoulder all need to be converted to MPa.
        Multipliers can be specified if a conversion from kPa to MPa is required. Optional keyword arguments for the
        `read_excel` function in Pandas can be specified as `**kwargs`.

        :param path: Path to the Excel file
        :param z_key: Column key for depth. Optional, default=None when 'z [m]' is the column key.
        :param qc_key: Column key for cone tip resistance. Optional, default=None when 'qc [MPa]' is the column key.
        :param fs_key: Column key for sleeve friction. Optional, default=None when 'fs [MPa]' is the column key.
        :param u2_key: Column key for pore pressure at shoulder. Optional, default=None when 'u2 [MPa]' is the column key.
        :param push_key: Column key for the current push (for downhole PCPT). Optional, default=None for a continuous push.
        :param qc_multiplier: Multiplier applied on cone tip resistance to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param fs_multiplier: Multiplier applied on sleeve friction to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param u2_multiplier: Multiplier applied on pore pressure at shoulder to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param add_zero_row: Boolean determining whether a datapoint needs to be added at zero depth.
        :param kwargs: Optional keyword arguments for the read_excel function in Pandas (e.g. sheet_name, header, ...)
        :return: Sets the columns 'z [m]', 'qc [MPa]', 'fs [MPa]' and 'u2 [MPa]' of the `.data` attribute
        """

        try:
            self.data = pd.read_excel(path, **kwargs)
            if push_key is None:
                self.data.loc[:, "Push"] = 1
            self.rename_columns(z_key=z_key, qc_key=qc_key, fs_key=fs_key, u2_key=u2_key, push_key=push_key)
            self.convert_columns(qc_multiplier=qc_multiplier, fs_multiplier=fs_multiplier, u2_multiplier=u2_multiplier)
        except Exception as err:
            raise ValueError("Error during reading of PCPT data. Review the error message and try again. - %s" % (
                str(err)))
        try:
            if self.data["z [m]"].min() != 0 and add_zero_row:
                # Append a row with zero data for vertical effective stress calculation
                self.add_zerodepth_row()
            else:
                pass
        except Exception as err:
            raise ValueError("Error appending of row with zero depth to PCPT data. Review the error message and try again. - %s" % (
                str(err)))

        try:
            self.dropna_rows()
        except Exception as err:
            raise ValueError("Error during dropping of empty rows. Review the error message and try again - %s" % str(
                err))

    def load_pandas(self, df, z_key=None, qc_key=None, fs_key=None, u2_key=None, push_key=None,
                    qc_multiplier=1, fs_multiplier=1, u2_multiplier=1, add_zero_row=True, ):
        """
        Loads PCPT from a Pandas dataframe. Specific column keys have to be provided for z, qc, fs and u2.
        If column keys are not specified, the following keys are used:

            - 'z [m]' for depth below mudline
            - 'qc [MPa]' for cone tip resistance
            - 'fs [MPa]' for sleeve friction
            - 'u2 [MPa]' for pore pressure at cone shoulder

        Note that cone tip resistance, sleeve friction and pore pressure at the shoulder all need to be converted to MPa

        :param df: Pandas dataframe with required column keys
        :param z_key: Column key for depth. Optional, default=None when 'z [m]' is the column key.
        :param qc_key: Column key for cone tip resistance. Optional, default=None when 'qc [MPa]' is the column key.
        :param fs_key: Column key for sleeve friction. Optional, default=None when 'fs [MPa]' is the column key.
        :param u2_key: Column key for pore pressure at shoulder. Optional, default=None when 'u2 [MPa]' is the column key.
        :param push_key: Column key for the current push (for downhole PCPT). Optional, default=None for a continuous push.
        :param qc_multiplier: Multiplier applied on cone tip resistance to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param fs_multiplier: Multiplier applied on sleeve friction to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param u2_multiplier: Multiplier applied on pore pressure at shoulder to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param add_zero_row: Boolean determining whether a datapoint needs to be added at zero depth.
        :return: Sets the columns 'z [m]', 'qc [MPa]', 'fs [MPa]' and 'u2 [MPa]' of the `.data` attribute
        """

        try:
            self.data = df
            if push_key is None:
                self.data.loc[:, "Push"] = 1
            self.rename_columns(z_key=z_key, qc_key=qc_key, fs_key=fs_key, u2_key=u2_key, push_key=push_key)
            self.convert_columns(qc_multiplier=qc_multiplier, fs_multiplier=fs_multiplier, u2_multiplier=u2_multiplier)
        except Exception as err:
            raise ValueError("Error during reading of PCPT data. Review the error message and try again. - %s" % (
                str(err)))
        try:
            if self.data["z [m]"].min() != 0 and add_zero_row:
                # Append a row with zero data for vertical effective stress calculation
                self.add_zerodepth_row()
            else:
                pass
        except Exception as err:
            raise ValueError("Error appending of row with zero depth to PCPT data. Review the error message and try again. - %s" % (
                str(err)))

        try:
            self.dropna_rows()
        except Exception as err:
            raise ValueError("Error during dropping of empty rows. Review the error message and try again - %s" % str(
                err))

    def load_asc(self, path, column_widths=[], skiprows=None, custom_headers=None,
                 z_key=None, qc_key=None, fs_key=None, u2_key=None, push_key=None,
                 qc_multiplier=1, fs_multiplier=1, u2_multiplier=1, add_zero_row=True, **kwargs):
        """
        Reads PCPT data from a Uniplot .asc file
        The widths of the columns for the .asc file need to be specified.

        :param path: Path to the .asc file
        :param column_widths: Column widths to use for the import (compulsory)
        :param skiprows: Number of rows to skip (optional, default=None for auto-detection, depends on the occurence of 'Data table' in the file)
        :param custom_headers: Custom headers to be used (default=None for auto-detection)
        :param z_key: Column key for depth. Optional, default=None when 'z [m]' is the column key.
        :param qc_key: Column key for cone tip resistance. Optional, default=None when 'qc [MPa]' is the column key.
        :param fs_key: Column key for sleeve friction. Optional, default=None when 'fs [MPa]' is the column key.
        :param u2_key: Column key for pore pressure at shoulder. Optional, default=None when 'u2 [MPa]' is the column key.
        :param push_key: Column key for the current push (for downhole PCPT). Optional, default=None for a continuous push.
        :param qc_multiplier: Multiplier applied on cone tip resistance to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param fs_multiplier: Multiplier applied on sleeve friction to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param u2_multiplier: Multiplier applied on pore pressure at shoulder to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param add_zero_row: Boolean determining whether a datapoint needs to be added at zero depth.
        :param kwargs: Optional keyword arguments for reading the datafile
        :return:
        """

        # Automatic detection of the number of rows to skip
        if skiprows is None:
            with open(path) as f:
                raw_file = pd.DataFrame(f.readlines(), columns=["asc lines"])
            # Find the line where the data starts
            start_line = raw_file[raw_file["asc lines"].str.contains('Data table')]
            skiprows = start_line.index[0] + 1
        else:
            skiprows = skiprows

        # Read the data
        self.data = pd.read_fwf(
            path,
            widths=column_widths,
            skiprows=skiprows, **kwargs)

        # Convert headers using the two rows with header information in the asc file
        if custom_headers is None:
            header_list = list(map(lambda key, unit: "%s [%s]" % (key, unit),
                list(self.data.columns), list(self.data.loc[0])))
        else:
            header_list = custom_headers

        new_headers = dict()
        for i, _head in enumerate(header_list):
            new_headers[self.data.columns[i]] = _head
        # Rename the headers and drop the first row
        self.data.rename(columns=new_headers, inplace=True)
        self.data = self.data.drop([0])
        self.data = self.data.astype('float')
        self.data.reset_index(drop=True, inplace=True)
        if push_key is None:
            self.data.loc[:, "Push"] = 1
        self.rename_columns(z_key=z_key, qc_key=qc_key, fs_key=fs_key, u2_key=u2_key, push_key=push_key)
        self.convert_columns(qc_multiplier=qc_multiplier, fs_multiplier=fs_multiplier, u2_multiplier=u2_multiplier)

        if self.data["z [m]"].min() != 0 and add_zero_row:
            # Append a row with zero data for vertical effective stress calculation
            self.add_zerodepth_row()
        else:
            pass

        try:
            self.dropna_rows()
        except Exception as err:
            raise ValueError("Error during dropping of empty rows. Review the error message and try again - %s" % str(
                err))

    def load_gef(self, path, headers, skiprows=None,
                 z_key=None, qc_key=None, fs_key=None, u2_key=None, push_key=None,
                 qc_multiplier=1, fs_multiplier=1, u2_multiplier=1, add_zero_row=True, **kwargs):
        """
        Reads PCPT data from a Uniplot .asc file
        The widths of the columns for the .asc file need to be specified.

        :param path: Path to the .asc file
        :param column_widths: Column widths to use for the import (compulsory)
        :param headers: Column headers to be used (compulsory)
        :param skiprows: Number of rows to skip (optional, default=None for auto-detection, depends on the occurence of 'Data table' in the file)
        :param z_key: Column key for depth. Optional, default=None when 'z [m]' is the column key.
        :param qc_key: Column key for cone tip resistance. Optional, default=None when 'qc [MPa]' is the column key.
        :param fs_key: Column key for sleeve friction. Optional, default=None when 'fs [MPa]' is the column key.
        :param u2_key: Column key for pore pressure at shoulder. Optional, default=None when 'u2 [MPa]' is the column key.
        :param push_key: Column key for the current push (for downhole PCPT). Optional, default=None for a continuous push.
        :param qc_multiplier: Multiplier applied on cone tip resistance to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param fs_multiplier: Multiplier applied on sleeve friction to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param u2_multiplier: Multiplier applied on pore pressure at shoulder to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param add_zero_row: Boolean determining whether a datapoint needs to be added at zero depth.
        :param kwargs: Optional keyword arguments for reading the datafile
        :return:
        """

        # Automatic detection of the number of rows to skip
        if skiprows is None:
            with open(path, 'rb') as f:
                raw_file = pd.DataFrame(f.readlines(), columns=["GEF lines"])
            # Find the line where the data starts
            start_line = raw_file[raw_file["GEF lines"].str.contains(b'#EOH=')]
            skiprows = start_line.index[0] + 1
        else:
            skiprows = skiprows

        # Read the data
        self.data = pd.read_csv(
            path,
            names=headers,
            skiprows=skiprows,
            delimiter=' ',
            **kwargs)

        self.data = self.data.astype('float')
        self.data.reset_index(drop=True, inplace=True)
        if push_key is None:
            self.data.loc[:, "Push"] = 1
        self.rename_columns(z_key=z_key, qc_key=qc_key, fs_key=fs_key, u2_key=u2_key, push_key=push_key)
        self.convert_columns(qc_multiplier=qc_multiplier, fs_multiplier=fs_multiplier, u2_multiplier=u2_multiplier)

        if self.data["z [m]"].min() != 0 and add_zero_row:
            # Append a row with zero data for vertical effective stress calculation
            self.add_zerodepth_row()
        else:
            pass

        try:
            self.dropna_rows()
        except Exception as err:
            raise ValueError("Error during dropping of empty rows. Review the error message and try again - %s" % str(
                err))

    def load_multi_asc(self, path, column_widths=[], skiprows=None, custom_headers=None,
                 z_key=None, qc_key=None, fs_key=None, u2_key=None, push_key=None,
                 qc_multiplier=1, fs_multiplier=1, u2_multiplier=1, add_zero_row=True,
                 start_string='Data table', end_string='UNICAS data file', **kwargs):
        """
        Reads PCPT data from a Uniplot .asc file with multiple pushes
        The widths of the columns for the .asc file need to be specified.

        :param path: Path to the .asc file
        :param column_widths: Column widths to use for the import (compulsory)
        :param skiprows: Number of rows to skip (optional, default=None for auto-detection, depends on the occurence of 'Data table' in the file)
        :param custom_headers: Custom headers to be used (default=None for auto-detection)
        :param z_key: Column key for depth. Optional, default=None when 'z [m]' is the column key.
        :param qc_key: Column key for cone tip resistance. Optional, default=None when 'qc [MPa]' is the column key.
        :param fs_key: Column key for sleeve friction. Optional, default=None when 'fs [MPa]' is the column key.
        :param u2_key: Column key for pore pressure at shoulder. Optional, default=None when 'u2 [MPa]' is the column key.
        :param push_key: Column key for the current push (for downhole PCPT). Optional, default=None for a continuous push.
        :param qc_multiplier: Multiplier applied on cone tip resistance to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param fs_multiplier: Multiplier applied on sleeve friction to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param u2_multiplier: Multiplier applied on pore pressure at shoulder to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param add_zero_row: Boolean determining whether a datapoint needs to be added at zero depth.
        :param start_string: String showing the start of the data
        :param end_string: String showing the end of the data
        :param kwargs: Optional keyword arguments for reading the datafile
        :return:
        """
        # Automatic detection of the number of rows to skip
        with open(path) as f:
            raw_file = pd.DataFrame(f.readlines(), columns=["asc lines"])
        start_lines = list(raw_file[raw_file["asc lines"].str.contains("Data table")].index)
        end_lines = list(raw_file[raw_file["asc lines"].str.startswith(" UNICAS data")].index)[1:]
        end_lines.append(raw_file.__len__())

        self.data = pd.DataFrame()

        colspecs = []
        colpos = 0
        for _w in column_widths:
            colspecs.append((colpos, colpos + _w))
            colpos += _w

        # Read the data
        for i, _start in enumerate(start_lines):

            push_data =  pd.read_fwf(
                path,
                colspecs=colspecs,
                skiprows=start_lines[i]+1,
                nrows=end_lines[i] - start_lines[i] - 3,
                **kwargs)
            if i == 0:
                header_list = list(map(lambda key, unit: "%s [%s]" % (key, unit),
                                       list(push_data.columns), list(push_data.loc[0])))
                new_headers = dict()
                for j, _head in enumerate(header_list):
                    new_headers[push_data.columns[j]] = _head

            push_data.rename(columns=new_headers, inplace=True)
            push_data.reset_index(drop=True, inplace=True)
            push_data = push_data.drop([0])
            push_data = push_data.astype('float')
            push_data.reset_index(drop=True, inplace=True)
            push_data.loc[:, 'Push'] = i+1
            self.data = pd.concat([self.data, push_data]).reset_index(drop=True)

        self.rename_columns(z_key=z_key, qc_key=qc_key, fs_key=fs_key, u2_key=u2_key, push_key=push_key)
        self.convert_columns(qc_multiplier=qc_multiplier, fs_multiplier=fs_multiplier, u2_multiplier=u2_multiplier)

        if self.data["z [m]"].min() != 0 and add_zero_row:
            # Append a row with zero data for vertical effective stress calculation
            self.add_zerodepth_row()
        else:
            pass

        try:
            self.dropna_rows()
        except Exception as err:
            raise ValueError("Error during dropping of empty rows. Review the error message and try again - %s" % str(
                err))


    def load_a00(self, path, column_widths=[], skiprows=None, custom_headers=None,
                 z_key=None, qc_key=None, fs_key=None, u2_key=None, push_key=None,
                 qc_multiplier=1, fs_multiplier=1, u2_multiplier=1, add_zero_row=True, **kwargs):
        """
        Reads PCPT data from a Uniplot .a00 file
        The widths of the columns for the .a00 file need to be specified.

        :param path: Path to the .a00 file
        :param column_widths: Column widths to use for the import (compulsory)
        :param skiprows: Number of rows to skip (optional, default=None for auto-detection, depends on the occurence of 'Data table' in the file)
        :param custom_headers: Custom headers to be used (default=None for auto-detection)
        :param z_key: Column key for depth. Optional, default=None when 'z [m]' is the column key.
        :param qc_key: Column key for cone tip resistance. Optional, default=None when 'qc [MPa]' is the column key.
        :param fs_key: Column key for sleeve friction. Optional, default=None when 'fs [MPa]' is the column key.
        :param u2_key: Column key for pore pressure at shoulder. Optional, default=None when 'u2 [MPa]' is the column key.
        :param push_key: Column key for the current push (for downhole PCPT). Optional, default=None for a continuous push.
        :param qc_multiplier: Multiplier applied on cone tip resistance to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param fs_multiplier: Multiplier applied on sleeve friction to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param u2_multiplier: Multiplier applied on pore pressure at shoulder to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param add_zero_row: Boolean determining whether a datapoint needs to be added at zero depth.
        :param kwargs: Optional keyword arguments for reading the datafile
        :return:
        """

        # Automatic detection of the number of rows to skip
        if skiprows is None:
            with open(path) as f:
                raw_file = pd.DataFrame(f.readlines(), columns=["a00 lines"])
            # Find the line where the data starts
            start_line = raw_file[raw_file["a00 lines"].str.contains('Reading')]
            skiprows = start_line.index[0]
        else:
            skiprows = skiprows

        # Read the data
        self.data = pd.read_fwf(
            path,
            widths=column_widths,
            skiprows=skiprows, **kwargs)

        # Convert headers using the two rows with header information in the asc file
        if custom_headers is None:
            header_list = list(map(lambda key, unit: "%s [%s]" % (key, unit),
                list(self.data.columns), list(self.data.loc[0])))
        else:
            header_list = custom_headers

        new_headers = dict()
        for i, _head in enumerate(header_list):
            new_headers[self.data.columns[i]] = _head
        # Rename the headers and drop the first row
        self.data.rename(columns=new_headers, inplace=True)
        self.data = self.data.drop([0, 1])
        self.data = self.data.astype('float')
        self.data.reset_index(drop=True, inplace=True)
        if push_key is None:
            self.data.loc[:, "Push"] = 1
        self.rename_columns(z_key=z_key, qc_key=qc_key, fs_key=fs_key, u2_key=u2_key, push_key=push_key)
        self.convert_columns(qc_multiplier=qc_multiplier, fs_multiplier=fs_multiplier, u2_multiplier=u2_multiplier)

        if self.data["z [m]"].min() != 0 and add_zero_row:
            # Append a row with zero data for vertical effective stress calculation
            self.add_zerodepth_row()
        else:
            pass

        try:
            self.dropna_rows()
        except Exception as err:
            raise ValueError("Error during dropping of empty rows. Review the error message and try again - %s" % str(
                err))

    def load_multiple_asc(self, folder, column_widths=[], skiprows=None, custom_headers=None,
                 z_key=None, qc_key=None, fs_key=None, u2_key=None, push_key=None,
                 qc_multiplier=1, fs_multiplier=1, u2_multiplier=1, **kwargs):
        """
        A PCPT can be provided as multiple .asc files in one folder. This method loops over the individual
        files and creates a combined `data` attribute with PCPT data. The method assumes that all .asc files have
        the same format.

        :param folder: Folder with the .asc files
        :param column_widths: Column widths to use for the import (compulsory)
        :param skiprows: Number of rows to skip (optional, default=None for auto-detection, depends on the occurence of 'Data table' in the file)
        :param custom_headers: Custom headers to be used (default=None for auto-detection)
        :param z_key: Column key for depth. Optional, default=None when 'z [m]' is the column key.
        :param qc_key: Column key for cone tip resistance. Optional, default=None when 'qc [MPa]' is the column key.
        :param fs_key: Column key for sleeve friction. Optional, default=None when 'fs [MPa]' is the column key.
        :param u2_key: Column key for pore pressure at shoulder. Optional, default=None when 'u2 [MPa]' is the column key.
        :param push_key: Column key for the current push (for downhole PCPT). Optional, default=None for a continuous push.
        :param qc_multiplier: Multiplier applied on cone tip resistance to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param fs_multiplier: Multiplier applied on sleeve friction to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param u2_multiplier: Multiplier applied on pore pressure at shoulder to convert to MPa (e.g. 0.001 to convert from kPa to MPa)
        :param kwargs: Optional keyword arguments for reading the datafiles
        :return: Sets the `data` attribute of the PCPTProcessing object
        """
        for i, filename in enumerate(os.listdir(folder)):
            if filename.endswith('.asc') or filename.endswith('.ASC'):
                if i > 0:
                    previous_data = deepcopy(self.data)
                    add_zero_row = False
                else:
                    add_zero_row = True
                self.load_asc(
                    self, filename, column_widths=column_widths, skiprows=skiprows,
                    custom_headers=custom_headers, z_key=z_key, qc_key=qc_key, fs_key=fs_key,
                    u2_key=u2_key, push_key=push_key,
                    qc_multiplier=qc_multiplier, fs_multiplier=fs_multiplier, u2_multiplier=u2_multiplier,
                    add_zero_row=add_zero_row, **kwargs)
                self.data.loc[:, "Push"] = i+1
                self.data = pd.concat([previous_data, self.data])
            else:
                pass

        self.data.sort_values('z [m]', inplace=True)
        self.data.reset_index(drop=True, inplace=True)

    def load_pydov(self, name, push_key=None, add_zero_row=True, **kwargs):
        """
        Load CPT data from Databank Ondergrond Vlaanderen based on the unique CPT name which can be found in DOV
        :param name: Unique identifier of the CPT in pydov
        :return: Sets the `data` attribute of the PCPTProcessing object
        """
        sondering = SonderingSearch()
        query = PropertyIsEqualTo(propertyname='sondeernummer',
                                  literal=name)
        self.data = sondering.search(query=query)
        if push_key is None:
            self.data.loc[:, "Push"] = 1
        self.rename_columns(z_key='z', qc_key='qc', fs_key='fs', u2_key='u')
        self.convert_columns(qc_multiplier=1, fs_multiplier=0.001, u2_multiplier=0.001)

        try:
            if self.data["z [m]"].min() != 0 and add_zero_row:
                # Append a row with zero data for vertical effective stress calculation
                self.add_zerodepth_row()
            else:
                pass
        except Exception as err:
            raise ValueError("Error appending of row with zero depth to PCPT data. Review the error message and try again. - %s" % (
                str(err)))

        try:
            self.dropna_rows()
        except Exception as err:
            raise ValueError("Error during dropping of empty rows. Review the error message and try again - %s" % str(
                err))

    def combine_pcpt(self, obj, keep="first"):
        """
        Combine PCPT data for two PCPTProcessing objects. The data of the second PCPT (`obj`) will be merged
        with the current object and the `data` attribute will be a combined dataframe.
        Only columns 'z [m]', 'qc [MPa]', 'fs [MPa]', 'u2 [MPa]' and 'Push' will be retained.

        :param obj: PCPTProcessing object containing the second PCPT info
        :param keep: Determines what to do in the area with overlap ("first", "second" or "both"). If "first" is chosen (default), the data from the first PCPT `self` is used in the area with overlap. For "second", data for the second PCPT `obj` is retained. If 'both' is selected, overlapping data will exist which might lead to presentation problems.
        :return: Updates the `data` attribute of the current PCPTProcessing object
        """
        if type(obj) != PCPTProcessing:
            raise TypeError("obj should be a PCPTProcessing object. Instead a %s was provided." % (type(obj)))

        z_min_first = self.data['z [m]'].min()
        z_max_first = self.data['z [m]'].max()
        z_min_second = obj.data['z [m]'].min()
        z_max_second = obj.data['z [m]'].max()

        self.data['Push'] = list(map(lambda _push: "1-%s" % _push, self.data['Push']))
        obj.data['Push'] = list(map(lambda _push: "2-%s" % _push, obj.data['Push']))

        # Second PCPT above first one
        if (z_max_second > z_min_first) and (z_min_second < z_min_first):
            if keep == 'first':
                first_data = self.data
                second_data = obj.data[obj.data["z [m]"] < z_min_first]
            elif keep == "second":
                first_data = self.data[self.data["z [m]"] > z_max_second]
                second_data = obj.data
            elif keep == "both":
                first_data = self.data
                second_data = obj.data
            else:
                raise ValueError("Invalid option for argument 'keep'. Choose from 'first', 'second' or 'both'")
        else:
            pass

        # Second PCPT below first one
        if (z_min_second < z_max_first) and (z_max_second > z_max_first):
            if keep == 'first':
                first_data = self.data
                second_data = obj.data[obj.data["z [m]"] > z_max_first]
            elif keep == "second":
                first_data = self.data[self.data["z [m]"] < z_min_second]
                second_data = obj.data
            elif keep == "both":
                first_data = self.data
                second_data = obj.data
            else:
                raise ValueError("Invalid option for argument 'keep'. Choose from 'first', 'second' or 'both'")
        else:
            pass

        # Combine data
        self.data = pd.concat([first_data, second_data])
        self.data.sort_values('z [m]', inplace=True)
        self.data.reset_index(drop=True, inplace=True)

    # endregion

    # region Layer-based processing and correction

    def map_properties(self, layer_profile, cone_profile=DEFAULT_CONE_PROPERTIES, vertical_total_stress=None, vertical_effective_stress=None, waterlevel=0):
        """
        Maps the soil properties defined in the layering and the cone properties to the grid
        defined by the cone data. The procedure also calculates the total and effective vertical stress.
        Note that pre-calculated arrays with total and effective vertical stress can also be supplied to the routine.
        These needs to have the same length as the array with PCPT depth data.

        :param layer_profile: ``SoilProfile`` object with the layer properties (need to contain the soil parameter ``Total unit weight [kN/m3]``
        :param cone_profile: ``SoilProfile`` object with the cone properties (default=``DEFAULT_CONE_PROPERTIES``)
        :param vertical_total_stress: Pre-calculated total vertical stress at PCPT depth nodes (default=None which will lead to calculation of total stress inside the routine)
        :param vertical_effective_stress: Pre-calculated effective vertical stress at PCPT depth nodes (default=None which will lead to calculation of total stress inside the routine)
        :param map_cone: Boolean determining whether cone properties need to be mapped or not (default=True)
        :return: Expands the dataframe `.data` with additional columns for the cone and soil properties
        """
        self.waterlevel = waterlevel
        self.layerdata = layer_profile

        # Validation
        if 'Total unit weight [kN/m3]' not in layer_profile.numerical_soil_parameters():
            raise ValueError("Soil layering profile needs to contain the parameter 'Total unit weight [kN/m3]'")

        # Map cone properties
        _mapped_cone_props = cone_profile.map_soilprofile(self.data['z [m]'])

        # Map layer properties
        for i, row in layer_profile.iterrows():
            layer_profile.loc[i, "Layer no"] = i+1
        _mapped_layer_props = layer_profile.map_soilprofile(self.data['z [m]'])

        # Join values to the CPT data
        self.data = self.data.join(_mapped_layer_props.set_index('z [m]'), on='z [m]', lsuffix='_left')
        self.data = self.data.join(_mapped_cone_props.set_index('z [m]'), on='z [m]', lsuffix='_left')

        # Calculation of total and effective vertical stress
        if vertical_total_stress is None:
            # Calculate vertical total stress
            self.data["Vertical total stress [kPa]"] = np.append(
                0,
                (np.array(self.data["z [m]"].diff()[1:]) *
                 np.array(self.data["Total unit weight [kN/m3]"][0:-1])).cumsum())
        else:
            self.data["Vertical total stress [kPa]"] = vertical_total_stress

        if vertical_effective_stress is None:
            # Calculation of vertical effective stress based on waterlevel
            self.data["Water pressure [kPa]"] = np.array(
                self.waterunitweight * (self.data["z [m]"] - self.waterlevel)).clip(min=0)
            self.data["Vertical effective stress [kPa]"] = self.data["Vertical total stress [kPa]"] - \
                                                           self.data["Water pressure [kPa]"]
        else:
            self.data["Vertical effective stress [kPa]"] = vertical_effective_stress

    def downhole_pcpt_corrections(self, area_ratio_override=np.nan):
        """
        Correct the PCPT for downhole effects. Select each push, find the start depth of the push and

        .. math::
            q_c = q_c^* + d \\cdot a \\cdot \\gamma_w

            u_2 = u_2^* + \\gamma_w \\cdot d

        :return:
        """

        if not self.downhole_corrected:
            try:
                self.data['qc raw [MPa]'] = self.data['qc [MPa]']
                self.data['u2 raw [MPa]'] = self.data['u2 [MPa]']
                for _push in self.data['Push'].unique():

                    push_data = self.data[self.data['Push'] == _push]
                    _push_z_min = push_data["z [m]"].min()
                    for i, row in push_data.iterrows():
                        if np.math.isnan(area_ratio_override):
                            area_ratio = row['area ratio [-]']
                        else:
                            area_ratio = area_ratio_override
                        self.data.loc[i, "qc [MPa]"] = row['qc raw [MPa]'] + \
                            0.001 * _push_z_min * area_ratio * self.waterunitweight
                        self.data.loc[i, "u2 [MPa]"] = row['u2 raw [MPa]'] + \
                            0.001 * self.waterunitweight * _push_z_min
                self.downhole_corrected = True
            except Exception as err:
                raise ValueError("Error during application of downhole corrections - %s" % str(err))
        else:
            warnings.warn("Downhole corrections have already been applied")

    def normalise_pcpt(self):
        """
        Carries out the necessary normalisation and correction on PCPT data to allow calculation of derived parameters and soil type classification.

        First, the cone resistance is corrected for the unequal area effect using the cone area ratio. The correction for total sleeve friction is not included as it is more uncommon. The procedure assumes that the pore pressure are measured at the shoulder of the cone. If this is not the case, corrections can be used which are not included in this function.

        During normalisation, the friction ratio and pore pressure ratio are calculated. Note that the total cone resistance is used for the friction ratio and pore pressure ratio calculation, the pore pressure ratio calculation also used the total vertical effective stress. The normalised cone resistance and normalised friction ratio are also calculated.

        Finally the net cone resistance is calculated.

        Note that the absence of pore water pressure measurements will lead to NaN values. Reasoning can be used (e.g. presence of rapidly draining layers) to edit the pore pressure data before running this method.

        .. math::
            q_c = q_c^* + d \\cdot a \\cdot \\gamma_w

            q_t = q_c + u_2 \\cdot (1 - a)

            u_2 = u_2^* + \\gamma_w \\cdot d

            \\Delta u_2 = u_2 - u_o

            R_f = \\frac{f_s}{q_t}

            B_q = \\frac{\\Delta u_2}{q_t - \\sigma_{vo}}

            Q_t = \\frac{q_t - \\sigma_{vo}}{\\sigma_{vo}^{\\prime}}

            F_r = \\frac{f_s}{q_t - \\sigma_{vo}}

            q_{net} = q_t - \\sigma_{vo}

        :return: Supplements the PCPT data (`.data`) with the normalised properties (column keys 'qt [MPa]', 'Delta u2 [MPa]', 'Rf [%]', 'Bq [-]', 'Qt [-]', 'Fr [%]', 'qnet [MPa]'
        """
        try:
            self.data['qt [MPa]'] = self.data['qc [MPa]'] + self.data['u2 [MPa]'] * (1 - self.data['area ratio [-]'])
            self.data['Delta u2 [MPa]'] = self.data['u2 [MPa]'] - 0.001 * self.data["Water pressure [kPa]"]
            self.data['Rf [%]'] = 100 * self.data['fs [MPa]'] / self.data['qt [MPa]']
            self.data['Bq [-]'] = self.data['Delta u2 [MPa]'] / (
                    self.data['qt [MPa]'] - 0.001 * self.data["Vertical total stress [kPa]"])
            self.data['Qt [-]'] = (self.data['qt [MPa]'] - 0.001 * self.data["Vertical total stress [kPa]"]) / (
                    0.001 * self.data["Vertical effective stress [kPa]"])
            self.data['Fr [%]'] = 100 * self.data['fs [MPa]'] / (
                    self.data['qt [MPa]'] - 0.001 * self.data["Vertical total stress [kPa]"])
            self.data['qnet [MPa]'] = self.data['qt [MPa]'] - 0.001 * self.data["Vertical total stress [kPa]"]
        except Exception as err:
            raise ValueError("Error during calculation of normalised properties."
                             "Review the error message and try again (%s)" % (str(err)))

    # endregion

    # region Data plotting

    def plot_raw_pcpt(self, qc_range=(0, 100), qc_tick=10, fs_range=(0, 1), fs_tick=0.1,
                      u2_range=(-0.1, 0.5), u2_tick=0.05, z_range=None, z_tick=2,
                      show_hydrostatic=True,
                      plot_height=700, plot_width=1000, return_fig=False,
                      plot_title=None, plot_margin=dict(t=100, l=50, b=50), color=None,
                      hydrostaticcolor=None, show_hydrostatic_legend=False, plot_layers=True):
        """
        Plots the raw PCPT data using the Plotly package. This generates an interactive plot.

        :param qc_range: Range for the cone tip resistance (default=(0, 100MPa))
        :param qc_tick: Tick interval for the cone tip resistance (default=10MPa)
        :param fs_range: Range for the sleeve friction (default=(0, 1MPa))
        :param fs_tick: Tick interval for sleeve friction (default=0.1MPa)
        :param u2_range: Range for the pore pressure at the shoulder (default=(-0.1, 0.5MPa))
        :param u2_tick: Tick interval for the pore pressure at the shoulder (default=0.05MPa)
        :param z_range: Range for the depth (default=None for plotting from zero to maximum cone penetration)
        :param z_tick: Tick interval for depth (default=2m)
        :param show_hydrostatic: Boolean determining whether hydrostatic pressure is shown on the pore pressure plot panel
        :param plot_height: Height for the plot (default=700px)
        :param plot_width: Width for the plot (default=1000px)
        :param return_fig: Boolean determining whether the figure needs to be returned (True) or plotted (False)
        :param plot_title: Plot for the title (default=None)
        :param plot_margin: Margin for the plot (default=dict(t=100, l=50, b=50))
        :param color: Color to be used for plotting (default=None for default plotly colors)
        :param color: Color to be used for plotting the hydrostatic pressure (default=None for default plotly colors)
        :param show_hydrostatic_legend: Boolean determining whether to show the hydrostatic pressure in the legend
        :return:
        """

        if z_range is None:
            z_range = (self.data['z [m]'].max(), 0)
        if color is None:
            color = DEFAULT_PLOTLY_COLORS[0]
        if hydrostaticcolor is None:
            hydrostaticcolor = DEFAULT_PLOTLY_COLORS[1]

        fig = subplots.make_subplots(rows=1, cols=3, print_grid=False, shared_yaxes=True)

        for _push in self.data["Push"].unique():
            push_data = self.data[self.data["Push"] == _push]
            trace1 = go.Scatter(
                x=push_data['qc [MPa]'],
                y=push_data['z [m]'],
                line=dict(color=color),
                showlegend=False, mode='lines', name=r'$ q_c $')
            fig.append_trace(trace1, 1, 1)
            trace2 = go.Scatter(
                x=push_data['fs [MPa]'],
                y=push_data['z [m]'],
                line=dict(color=color),
                showlegend=False, mode='lines', name=r'$ f_s $')
            fig.append_trace(trace2, 1, 2)
            trace3 = go.Scatter(
                x=push_data['u2 [MPa]'],
                y=push_data['z [m]'],
                line=dict(color=color),
                showlegend=False, mode='lines', name=r'$ u $')
            fig.append_trace(trace3, 1, 3)
        if show_hydrostatic:
            trace3 = go.Scatter(
                x=0.001 * self.data['z [m]'] * self.waterunitweight,
                y=self.data['z [m]'],
                line=dict(color=hydrostaticcolor, dash='dot'),
                showlegend=show_hydrostatic_legend, mode='lines', name=r'$ u_{hydrostatic} $')
            fig.append_trace(trace3, 1, 3)
        # Plot layers
        try:
            for i, row in self.layerdata.iterrows():
                if i > 0:
                    layer_trace_qc = go.Scatter(
                        x=qc_range,
                        y=(row[self.layerdata.depth_from_col], row[self.layerdata.depth_from_col]),
                        line=dict(color='black', dash='dot'),
                        showlegend=False, mode='lines')
                    fig.append_trace(layer_trace_qc, 1, 1)
                    layer_trace_fs = go.Scatter(
                        x=fs_range,
                        y=(row[self.layerdata.depth_from_col], row[self.layerdata.depth_from_col]),
                        line=dict(color='black', dash='dot'),
                        showlegend=False, mode='lines')
                    fig.append_trace(layer_trace_fs, 1, 2)
                    layer_trace_u2 = go.Scatter(
                        x=u2_range,
                        y=(row[self.layerdata.depth_from_col], row[self.layerdata.depth_from_col]),
                        line=dict(color='black', dash='dot'),
                        showlegend=False, mode='lines')
                    fig.append_trace(layer_trace_u2, 1, 3)
        except:
            pass
        fig['layout']['xaxis1'].update(
            title=r'$ q_c \ [\text{MPa}] $', side='top', anchor='y',
            range=qc_range, dtick=qc_tick)
        fig['layout']['xaxis2'].update(
            title=r'$ f_s \ [\text{MPa}]  $', side='top', anchor='y',
            range=fs_range, dtick=fs_tick)
        fig['layout']['xaxis3'].update(
            title=r'$ u  \ [\text{MPa}] $', side='top', anchor='y',
            range=u2_range, dtick=u2_tick)
        fig['layout']['yaxis1'].update(
            title=r'$ z \ [\text{m}]$', range=z_range, dtick=z_tick)
        fig['layout'].update(
            title=plot_title, hovermode='closest',
            height=plot_height, width=plot_width,
            margin=plot_margin)
        if return_fig:
            return fig
        else:
            iplot(fig, filename='pcptplot', config={'showLink': False})


    def plot_normalised_pcpt(
            self, qt_range=(0, 3), fr_range=(-1, 1),
            bq_range=(-0.6, 1.4), bq_tick=0.2, z_range=None, z_tick=2,
            plot_height=700, plot_width=1000, color=None, return_fig=False, plot_title=None):
        """
        Plots the normalised PCPT properties vs depth.

        :param qt_range: Range for Qt (optional, default is (0, 3) for 1 to 1000 on log scale)
        :param fr_range: Range for Fr (optional, default is (-1, 1) for 0.1 to 10 on log scale)
        :param bq_range: Range for Bq (optional, default is (-0.6, 1.4))
        :param bq_tick: Tick mark interval for Bq (optional, default=0.2)
        :param z_range: Range for depths (optional, default is (0, maximum PCPT depth)
        :param z_tick: Tick mark distance for PCPT (optional, default=2)
        :param plot_height: Height of the plot in pixels
        :param plot_width: Width of the plot in pixels
        :param return_fig: Boolean determining whether the figure is returned or the plot is generated; Default behaviour is to generate the plot.
        :param plot_title: Plot title (optional, default=None for no title)
        :return: Returns the figure if return_fig=True. Otherwise the plot is displayed.
        """
        if z_range is None:
            z_range = (self.data['z [m]'].max(), 0)
        if color is None:
            color = DEFAULT_PLOTLY_COLORS[0]

        fig = subplots.make_subplots(rows=1, cols=3, print_grid=False, shared_yaxes=True)
        trace1 = go.Scatter(
            x=self.data['Qt [-]'],
            y=self.data['z [m]'],
            line=dict(color=color),
            showlegend=False, mode='lines', name=r'$ Q_t $')
        fig.append_trace(trace1, 1, 1)
        trace2 = go.Scatter(
            x=self.data['Fr [%]'],
            y=self.data['z [m]'],
            line=dict(color=color),
            showlegend=False, mode='lines', name=r'$ F_r $')
        fig.append_trace(trace2, 1, 2)
        trace3 = go.Scatter(
            x=self.data['Bq [-]'],
            y=self.data['z [m]'],
            line=dict(color=color),
            showlegend=False, mode='lines', name=r'$ B_q $')
        fig.append_trace(trace3, 1, 3)
        # Plot layers
        try:
            for i, row in self.layerdata.iterrows():
                if i > 0:
                    layer_trace_qt = go.Scatter(
                        x=(10 ** qt_range[0], 10 ** qt_range[1]),
                        y=(row[self.layerdata.depth_from_col], row[self.layerdata.depth_from_col]),
                        line=dict(color='black', dash='dot'),
                        showlegend=False, mode='lines')
                    fig.append_trace(layer_trace_qt, 1, 1)
                    layer_trace_fr = go.Scatter(
                        x=(10 ** fr_range[0], 10 ** fr_range[1]),
                        y=(row[self.layerdata.depth_from_col], row[self.layerdata.depth_from_col]),
                        line=dict(color='black', dash='dot'),
                        showlegend=False, mode='lines')
                    fig.append_trace(layer_trace_fr, 1, 2)
                    layer_trace_bq = go.Scatter(
                        x=bq_range,
                        y=(row[self.layerdata.depth_from_col], row[self.layerdata.depth_from_col]),
                        line=dict(color='black', dash='dot'),
                        showlegend=False, mode='lines')
                    fig.append_trace(layer_trace_bq, 1, 3)
        except:
            pass
        fig['layout']['xaxis1'].update(
            title=r'$ Q_t \ [\text{-}] $', side='top', anchor='y', type='log',
            range=qt_range)
        fig['layout']['xaxis2'].update(
            title=r'$ F_r \ [\text{%}]  $', side='top', anchor='y', type='log',
            range=fr_range)
        fig['layout']['xaxis3'].update(
            title=r'$ B_q \ [\text{-}] $', side='top', anchor='y',
            range=bq_range, dtick=bq_tick)
        fig['layout']['yaxis1'].update(
            title=r'$ z \ [\text{m}]$', range=z_range, dtick=z_tick)
        fig['layout'].update(
            title=plot_title,
            height=plot_height, width=plot_width,
            margin=dict(t=100, l=50, b=50))
        if return_fig:
            return fig
        else:
            iplot(fig, filename='pcptnormalisedplot', config={'showLink': False})

    def plot_properties(
            self, prop_keys, plot_ranges, plot_ticks, z_range=None, z_tick=2,
            legend_titles=None, axis_titles=None,
            plot_height=700, plot_width=1000, colors=None, return_fig=False, plot_title=None):
        """
        Plots the soil and/or PCPT properties vs depth.

        :param prop_keys: Tuple of tuples with the keys to be plotted. Keys in the same tuple are plotted on the same panel
        :param plot_ranges: Tuple of tuples with ranges for the panels of the plot
        :param plot_ticks: Tuple with tick intervals for the plot panels
        :param z_range: Range for depths (optional, default is (0, maximum PCPT depth)
        :param z_tick: Tick mark distance for PCPT (optional, default=2)
        :param legend_titles: Tuple with entries to be used in the legend. If left blank, the keys are used
        :param axis_titles: Tuple with entries to be used as axis labels. If left blank, the keys are used
        :param plot_height: Height of the plot in pixels
        :param plot_width: Width of the plot in pixels
        :param return_fig: Boolean determining whether the figure is returned or the plot is generated; Default behaviour is to generate the plot.
        :param plot_title: Plot title (optional, default=None for no title)
        :return: Returns the figure if return_fig=True. Otherwise the plot is displayed.
        """
        if z_range is None:
            z_range = (self.data['z [m]'].max(), 0)
        if colors is None:
            colors = DEFAULT_PLOTLY_COLORS
        if legend_titles is None:
            legend_titles = prop_keys
        if axis_titles is None:
            axis_titles = prop_keys

        fig = subplots.make_subplots(rows=1, cols=prop_keys.__len__(), print_grid=False, shared_yaxes=True)
        for i, _props in enumerate(prop_keys):
            for j, _prop in enumerate(_props):
                trace = go.Scatter(
                    x=self.data[_prop],
                    y=self.data['z [m]'],
                    line=dict(color=colors[j]),
                    showlegend=False, mode='lines', name=legend_titles[i][j])
                fig.append_trace(trace, 1, i+1)
        # Plot layers
        try:
            for i, row in self.layerdata.iterrows():
                if i > 0:
                    for i, _range in enumerate(plot_ranges):
                        layer_trace = go.Scatter(
                            x=_range,
                            y=(row[self.layerdata.depth_from_col], row[self.layerdata.depth_from_col]),
                            line=dict(color='black', dash='dot'),
                            showlegend=False, mode='lines')
                        fig.append_trace(layer_trace, 1, i+1)
        except:
            pass
        for i, _range in enumerate(plot_ranges):
            fig['layout']['xaxis%i' % (i+1)].update(
                title=axis_titles[i], side='top', anchor='y',
                range=_range, dtick=plot_ticks[i])
        fig['layout']['yaxis1'].update(
            title=r'$ z \ [\text{m}]$', range=z_range, dtick=z_tick)
        fig['layout'].update(
            title=plot_title,
            height=plot_height, width=plot_width,
            margin=dict(t=100, l=50, b=50))
        if return_fig:
            return fig
        else:
            iplot(fig, filename='propertiesplot', config={'showLink': False})

    def plot_properties_withlog(self, prop_keys, plot_ranges, plot_ticks,
            legend_titles=None, axis_titles=None, **kwargs):
        """
        Plots CPT properties vs depth and includes a mini-log on the left-hand side.
        The minilog is composed based on the entries in the ``Soil type`` column of the layering
        :param prop_keys: Tuple of tuples with the keys to be plotted. Keys in the same tuple are plotted on the same panel
        :param plot_ranges: Tuple of tuples with ranges for the panels of the plot
        :param plot_ticks: Tuple with tick intervals for the plot panels
        :param z_range: Range for depths (optional, default is (0, maximum PCPT depth)
        :param z_tick: Tick mark distance for PCPT (optional, default=2)
        :param legend_titles: Tuple with entries to be used in the legend. If left blank, the keys are used
        :param axis_titles: Tuple with entries to be used as axis labels. If left blank, the keys are used
        :param **kwargs: Specify keyword arguments for the ``general.plotting.plot_with_log`` function
        :return: Plotly figure with mini-log
        """
        # Validate if 'Soil type' column is present in the layering
        if 'Soil type' not in self.layerdata.columns:
            raise ValueError("Layering should contain the column 'Soil type'")

        if legend_titles is None:
            legend_titles = prop_keys
        if axis_titles is None:
            axis_titles = prop_keys

        _x = []
        _z = []

        for _panel in prop_keys:
            _x_panel = []
            _z_panel = []
            for _prop in _panel:
                _x_panel.append(self.data[_prop])
                _z_panel.append(self.data['z [m]'])
            _x.append(_x_panel)
            _z.append(_z_panel)

        fig = plot_with_log(
            x=_x,
            z=_z,
            names=legend_titles,
            soildata=self.layerdata,
            xtitles=axis_titles,
            xranges=plot_ranges,
            dticks=plot_ticks,
            **kwargs
        )
        return fig

    def plot_robertson_chart(self, start_depth=None, end_depth=None,
                             qt_range=(0, 3), fr_range=(-1, 1),
            bq_range=(-0.6, 1.4), bq_tick=0.2,
            plot_height=700, plot_width=1000, return_fig=False, plot_title=None,
            backgroundimagedir=""):
        """
        Plots the normalised PCPT points in the Robertson chart to distinguish the soil type. The display can be limited
        to a specific depth range (by specifying `start_depth` and `end_depth`. The color coding is based on the layer.
        :return: Returns the figure if return_fig=True. Otherwise the plot is displayed.
        """

        if start_depth is None:
            start_depth = self.data["z [m]"].min()
        if end_depth is None:
            end_depth = self.data["z [m]"].max()

        selected_data = self.data[(self.data["z [m]"] >= start_depth) & (self.data["z [m]"] <= end_depth)]

        fig = subplots.make_subplots(rows=1, cols=2, print_grid=False)
        for i, layer in enumerate(selected_data["Layer no"].unique()):
            color = DEFAULT_PLOTLY_COLORS[i % DEFAULT_PLOTLY_COLORS.__len__()]
            layer_data = selected_data[selected_data["Layer no"] == layer]
            layer_min_depth = layer_data['z [m]'].min()
            layer_max_depth = layer_data['z [m]'].max()
            friction_trace = go.Scatter(
                x=layer_data['Fr [%]'],
                y=layer_data['Qt [-]'],
                showlegend=True,  # Don't show the legend
                mode='markers',
                name="Layer %i - %.2fm - %.2fm" % (layer, layer_min_depth, layer_max_depth),
                text=layer_data["z [m]"],
                marker=dict(size=10,  # Make markers transparent (last number is 0)
                            opacity=0.5,  # Add some opacity for better display
                            color=color,
                            line=dict(width=1, color=color)))  # Add a line around the markers
            fig.append_trace(friction_trace, 1, 1)
            pressure_trace = go.Scatter(
                x=layer_data['Bq [-]'],
                y=layer_data['Qt [-]'],
                showlegend=False,
                mode='markers',
                text=layer_data["z [m]"],
                marker=dict(size=10,   # Make markers transparent (last number is 0)
                            opacity=0.5,  # Add some opacity for better display
                            color=color,
                            line=dict(width=1, color=color)))  # Add a line around the markers
            fig.append_trace(pressure_trace, 1, 2)
        fig['layout']['xaxis1'].update(title=r'$ F_r \ [\%] $', range=fr_range, type='log')
        fig['layout']['yaxis1'].update(title=r'$ Q_t \ [-] $', range=qt_range, type='log')
        fig['layout']['xaxis2'].update(title=r'$ B_q \ [-] $', range=bq_range, dtick=bq_tick)
        fig['layout']['yaxis2'].update(title=r'$ Q_t \ [-] $', range=qt_range, type='log')
        fig['layout'].update(
            height=plot_height,
            width=plot_width,
            title=plot_title,
            hovermode='closest',
            images=[
                # Image for Qt-Fr relation
                dict(
                    source=os.path.join(backgroundimagedir, 'robertsonFr.png'),
                    xref='x',
                    yref='y',
                    x=-1,
                    y=3,
                    sizex=2,
                    sizey=3,
                    sizing='stretch',
                    opacity=0.7,
                    layer='below',
                ),
                # Image for Qt-Bq relation
                dict(
                    source=os.path.join(backgroundimagedir, 'robertsonBq.png'),
                    xref='x2',
                    yref='y2',
                    x=-0.6,
                    y=3,
                    sizex=2,
                    sizey=3,
                    sizing='stretch',
                    opacity=0.7,
                    layer='below',
                )
            ]
        )
        if return_fig:
            return fig
        else:
            iplot(fig, filename='robertson', config={'showLink': False})

    # endregion

    # region Correlations

    def apply_correlation(self, name, outkey, resultkey, apply_for_soiltypes='all', **kwargs):
        """
        Applies a correlation to the given PCPT data. The name of the correlation needs to be chosen from the following available correlations.
        Each correlation corresponds to a function in the `pcpt` module.
        By default, the correlation is applied to the entire depth range. However, a restriction on the soil
        types to which the correlation can be applied can be specified with the `apply_for_soiltypes` keyword argument.
        A list with the soil types for which the correlation needs to be applied can be provided.

            - Robertson and Wride (1998) (`behaviourindex_pcpt_robertsonwride`) - Calculation of soil behaviour type index from cone tip resistance and sleeve friction
            - Rix and Stokoe (1991) (`gmax_sand_rixstokoe`) - Calculation of small-strain shear modulus for uncemented silica sand from cone tip resistance and vertical effective stress
            - Mayne and Rix (1993) (`gmax_clay_maynerix`) - Calculation of small-strain shear modulus for clay from cone tip resistance
            - Baldi et al (1986) - NC sand (`relativedensity_ncsand_baldi`) - Calculation of relative density of normally consolidated silica sand
            - Baldi et al (1986) - OC sand (`relativedensity_ocsand_baldi`) - Calculation of relative density of overconsolidated silica sand
            - Jamiolkowski et al (2003) (`relativedensity_sand_jamiolkowski`) - Calculation of relative density dry and saturated silica sand
            - Kulhawy and Mayne (1990) (`frictionangle_sand_kulhawymayne`) - Calculation of effective friction angle for sand
            - Rad and Lunne (1988): (`undrainedshearstrength_clay_radlunne`) - Calculation of undrained shear strength for clay based on empirical cone factor Nk

        Note that certain correlations require either the calculation of normalised properties or application of preceding correlations

        :param name: Name of the correlation according to the list defined above
        :param outkey: Key used to the output column
        :param resultkey: Key of the output dictionary of the correlation to be used
        :param apply_for_soiltypes: List with soil types to which the correlation needs the be applied.
        :param kwargs: Optional keyword arguments for the correlation.
        :return: Adds a column with key `outkey` to the dataframe with PCPT data
        """
        self.data.rename(columns=PCPT_KEY_MAPPING, inplace=True)
        for i, row in self.data.iterrows():
            if apply_for_soiltypes == 'all' or row['Soil type'] in apply_for_soiltypes:
                params = merge_two_dicts(kwargs, dict(row))
                self.data.loc[i, outkey] = CORRELATIONS[name](**params)[resultkey]
            else:
                self.data.loc[i, outkey] = np.nan
        self.data.rename(columns=reverse_dict(PCPT_KEY_MAPPING), inplace=True)
    # endregion

    # region Design lines
    def load_design_profile(self, design_data):
        """
        Loads a design profile for the parameters available in the `data` attribute of the PCPTProcessing object.
        This design profile can then be plotted vs the PCPT data.

        :param design_data: Pandas dataframe with design soil profile. Linear variation of numerical parameters is expected so for example, use columns `Su from [kPa]` and `Su to [kPa]`
        :return: Sets the `designprofile` attribute of the PCPTProcessing object
        """
        self.designprofile = design_data

    def plot_design_profile(self, prop_keys, design_keys, plot_ranges, plot_ticks, z_range=None, z_tick=2,
            legend_titles=None, axis_titles=None,
            plot_height=700, plot_width=1000, colors=None, design_color='red', design_dash='dot',
                            return_fig=False, plot_title=None):
        """
        Plots the soil and/or PCPT properties vs depth.

        :param prop_keys: Tuple of tuples with the keys to be plotted. Keys in the same tuple are plotted on the same panel
        :param design_keys: Tuple of tuples with the design profile keys to be plotted. Keys in the same tuple are plotted on the same panel (note that `()` is an empty tuple). The design keys need to be specified as for example `Gmax [kPa]` if the `designprofile` attribute contains keys `Gmax from [kPa]` and `Gmax to [kPa]`
        :param plot_ranges: Tuple of tuples with ranges for the panels of the plot
        :param plot_ticks: Tuple with tick intervals for the plot panels
        :param z_range: Range for depths (optional, default is (0, maximum PCPT depth)
        :param z_tick: Tick mark distance for PCPT (optional, default=2)
        :param legend_titles: Tuple with entries to be used in the legend. If left blank, the keys are used
        :param axis_titles: Tuple with entries to be used as axis labels. If left blank, the keys are used
        :param plot_height: Height of the plot in pixels
        :param plot_width: Width of the plot in pixels
        :param colors: Color list to be used in the plotting panels
        :param design_color: Color to be used for the design lines
        :param design_dash: Dash style to be used for the design lines
        :param return_fig: Boolean determining whether the figure is returned or the plot is generated; Default behaviour is to generate the plot.
        :param plot_title: Plot title (optional, default=None for no title)
        :return: Returns the figure if return_fig=True. Otherwise the plot is displayed.
        """
        if z_range is None:
            z_range = (self.data['z [m]'].max(), 0)
        if colors is None:
            colors = DEFAULT_PLOTLY_COLORS
        if legend_titles is None:
            legend_titles = prop_keys
        if axis_titles is None:
            axis_titles = prop_keys

        fig = subplots.make_subplots(rows=1, cols=prop_keys.__len__(), print_grid=False, shared_yaxes=True)
        for i, _props in enumerate(prop_keys):
            for j, _prop in enumerate(_props):
                trace = go.Scatter(
                    x=self.data[_prop],
                    y=self.data['z [m]'],
                    line=dict(color=colors[j]),
                    showlegend=False, mode='lines', name=legend_titles[i][j])
                fig.append_trace(trace, 1, i+1)

        for i, _props in enumerate(design_keys):
            for j, _prop in enumerate(_props):
                try:
                    keybase = re.split(" \[", _prop)
                    from_key = "%s from [%s" % (keybase[0], keybase[1])
                    to_key = "%s to [%s" % (keybase[0], keybase[1])
                    design_trace = go.Scatter(
                        x=np.insert(
                            np.array(self.designprofile[to_key]),
                            np.arange(len(self.designprofile)),
                            np.array(self.designprofile[from_key])),
                        y=np.insert(
                            np.array(self.designprofile["z to [m]"]),
                            np.arange(len(self.designprofile)),
                            np.array(self.designprofile["z from [m]"])),
                        line=dict(color=design_color, dash=design_dash),
                        showlegend=False, mode='lines', name=keybase[0])
                    fig.append_trace(design_trace, 1, i+1)
                except:
                    pass

        # Plot layers
        try:
            for i, row in self.layerdata.iterrows():
                if i > 0:
                    for i, _range in enumerate(plot_ranges):
                        layer_trace = go.Scatter(
                            x=_range,
                            y=(row[self.layerdata.depth_from_col], row[self.layerdata.depth_from_col]),
                            line=dict(color='black', dash='dot'),
                            showlegend=False, mode='lines')
                        fig.append_trace(layer_trace, 1, i+1)
        except:
            pass
        for i, _range in enumerate(plot_ranges):
            fig['layout']['xaxis%i' % (i+1)].update(
                title=axis_titles[i], side='top', anchor='y',
                range=_range, dtick=plot_ticks[i])
        fig['layout']['yaxis1'].update(
            title=r'$ z \ [\text{m}]$', range=z_range, dtick=z_tick)
        fig['layout'].update(
            title=plot_title,
            height=plot_height, width=plot_width,
            margin=dict(t=100, l=50, b=50))
        if return_fig:
            return fig
        else:
            iplot(fig, filename='propertiesplot', config={'showLink': False})

    # endregion

    # region Data export

    def to_json(self, write_file=False, output_path=None):
        """
        Write the PCPT object to a JSON file. JSON can either be returned or a JSON file can be written.

        :param write_file: Boolean determining whether a file is written or not (default=False). If True, a file is written to `output_path`
        :param output_path: A valid path to the output .json file (include the file suffix).
        :return: If no file is returned, the JSON containing the location and data of the PCPT is returned.
        """
        dict_pcpt = {
            'easting': self.easting,
            'northing': self.northing,
            'waterdepth': self.elevation,
            'srid': self.srid,
            'data': self.data.to_json()
        }
        if write_file:
            with open(output_path, 'w') as outfile:
                json.dump(dict_pcpt, outfile)
        else:
            return json.dumps(dict_pcpt)

    # endregion



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Bruno Stuyts'

# Native Python packages
import unittest
import os

# 3rd party packages
import pandas as pd
import numpy as np

# Project imports
from groundhog.general import parameter_mapping


class Test_map_depth_properties(unittest.TestCase):

    def setUp(self):
        self.target_df = pd.DataFrame({'z [m]': np.linspace(0, 20, 21),})
        self.layering_df = pd.DataFrame({
            'z from [m]': np.array([0, 5, 10]),
            'z to [m]': np.array([5, 10, 20]),
            'Submerged unit weigth [kN/m3]': np.array([8.5, 7, 10]),
            'Undrained shear strength from [kPa]': np.array([np.nan, 50, np.nan]),
            'Undrained shear strength to [kPa]': np.array([np.nan, 60, np.nan]),
            'Effective friction angle from [deg]': np.array([32.5, np.nan, 40]),
            'Effective friction angle to [deg]': np.array([35, np.nan, 40]),
            'Soil type': np.array(['SAND', 'CLAY', 'SAND'])
        })
        self.target_df_altcol = pd.DataFrame({'Depth [m]': np.linspace(0, 20, 21), })
        self.layering_df_altcols = pd.DataFrame({
            'Depth from [m]': np.array([0, 5, 10]),
            'Depth to [m]': np.array([5, 10, 20]),
            'Submerged unit weigth [kN/m3]': np.array([8.5, 7, 10]),
            'Undrained shear strength from [kPa]': np.array([np.nan, 50, np.nan]),
            'Undrained shear strength to [kPa]': np.array([np.nan, 60, np.nan]),
            'Effective friction angle from [deg]': np.array([32.5, np.nan, 40]),
            'Effective friction angle to [deg]': np.array([35, np.nan, 40]),
            'Soil type': np.array(['SAND', 'CLAY', 'SAND'])
        })


    def test_map_depth_properties(self):
        result = parameter_mapping.map_depth_properties(
            target_df=self.target_df,
            layering_df=self.layering_df)
        self.assertEqual(
            result[result["z [m]"] == 2]["Submerged unit weigth [kN/m3]"].iloc[0], 8.5
        )
        self.assertEqual(
            result[result["z [m]"] == 5]["Soil type"].iloc[0], 'CLAY'
        )
        self.assertEqual(
            result[result["z [m]"] == 5]["Undrained shear strength [kPa]"].iloc[0], 50
        )
        self.assertTrue(
            np.math.isnan(result[result["z [m]"] == 5]["Effective friction angle [deg]"].iloc[0]))
        result = parameter_mapping.map_depth_properties(
            target_df=self.target_df_altcol,
            layering_df=self.layering_df_altcols,
            target_z_key="Depth [m]",
            layering_zfrom_key="Depth from [m]",
            layering_zto_key="Depth to [m]"
            )
        self.assertEqual(
            result[result["Depth [m]"] == 2]["Submerged unit weigth [kN/m3]"].iloc[0], 8.5
        )
        self.assertEqual(
            result[result["Depth [m]"] == 5]["Soil type"].iloc[0], 'CLAY'
        )
        self.assertEqual(
            result[result["Depth [m]"] == 5]["Undrained shear strength [kPa]"].iloc[0], 50
        )
        self.assertTrue(
            np.math.isnan(result[result["Depth [m]"] == 5]["Effective friction angle [deg]"].iloc[0])
        )

